from typing import List, Dict, Any, Union, Optional
import pydantic


class Tag(pydantic.BaseModel):
    start_tag: str
    end_tag: str
    content: Union[str, List["Tag"], "Tag"]
    parent: Optional["Tag"] = None


class TagExtractor:
    def __init__(self, text: str):
        self.text = text
        self.pos = -1
        self.len = len(text)
        self.root_tag = Tag(start_tag="<_ROOT_>", end_tag="</_ROOT_>", content=[])
        self.current_tag = None

    def peek(self) -> str:
        if self.pos + 1 < self.len:
            return self.text[self.pos + 1]
        return ""

    def peek2(self) -> str:
        if self.pos + 1 < self.len:
            return self.text[self.pos + 2]
        return ""

    def peek3(self) -> str:
        if self.pos + 1 < self.len:
            return self.text[self.pos + 3]
        return ""

    def next(self) -> str:
        if self.pos < self.len - 1:
            self.pos += 1
            char = self.text[self.pos]
            return char
        return ""

    def is_full_tag(self) -> bool:
        return self.current_tag.start_tag and self.current_tag.end_tag

    def is_start_tag(self) -> bool:
        return self.peek() and self.peek() == "<" and self.peek2() == "_"

    def extract_start_tag(self) -> str:
        tag = ""

        while self.peek() and self.peek() != ">":
            tag += self.next()
        tag += self.next()

        if self.current_tag is None or self.current_tag == self.root_tag:
            self.current_tag = Tag(
                start_tag=tag, end_tag="", content="", parent=self.root_tag
            )
            self.root_tag.content.append(self.current_tag)
        ## 当前tag已经闭合，找到当前tag的父tag,遇到新tag
        elif self.is_full_tag():
            parent = self.current_tag.parent or self.current_tag
            current_tag = Tag(
                start_tag=tag,
                end_tag="",
                content="",
                parent=parent,
            )
            if isinstance(parent.content, list):
                parent.content.append(current_tag)
            else:
                s = self.current_tag.content
                parent.content = []
                if s:
                    parent.content.append(s)
                parent.content.append(current_tag)
            self.current_tag = current_tag
        ## 当前tag还没有闭合，直接添加到当前tag的content中
        elif not self.is_full_tag():
            current_tag = Tag(
                start_tag=tag, end_tag="", content="", parent=self.current_tag
            )
            if isinstance(self.current_tag.content, list):
                self.current_tag.content.append(current_tag)
            else:
                s = self.current_tag.content
                self.current_tag.content = []
                if s:
                    self.current_tag.content.append(s)
                self.current_tag.content.append(current_tag)
            self.current_tag = current_tag

        return tag

    def is_end_tag(self) -> bool:
        return (
            self.peek()
            and self.peek() == "<"
            and self.peek2() == "/"
            and self.peek3() == "_"
        )

    def extract_end_tag(self) -> str:
        tag = ""
        while self.peek() and self.peek() != ">":
            tag += self.next()
        tag += self.next()
        self.current_tag.end_tag = tag
        self.current_tag = self.current_tag.parent
        return tag

    def consume_blank(self):
        while (
            self.peek() == "\n"
            or self.peek() == " "
            or self.peek() == "\t"
            or self.peek() == "\r"
        ):
            self.next()

    def is_in_tag(self) -> bool:
        return (
            self.current_tag
            and self.current_tag.start_tag
            and not self.current_tag.end_tag
        )

    def is_tag_content(self) -> bool:
        if not self.root_tag:
            return False
        temp_pos = self.pos
        self.consume_blank()
        if self.is_start_tag():
            self.pos = temp_pos            
            return True
        if self.is_end_tag():
            self.pos = temp_pos
            return True                
        return False

    def extract_str_content(self) -> str:
        content = ""
        while not self.is_start_tag() and not self.is_end_tag():
            content += self.next()
         
        self.current_tag.content = content
        return content

    def is_not_in_tag_str(self) -> bool:
        if not self.root_tag:
            return True
        if not self.current_tag:
            return True
        if not self.current_tag.start_tag and not self.current_tag.end_tag:
            return True
        if self.current_tag.start_tag and self.current_tag.end_tag:
            return True
        return False

    def extract_content_not_in_tag(self) -> str:
        content = ""
        while self.peek() and not self.is_start_tag() and not self.is_end_tag():
            content += self.next()
        return content

    def extract(self) -> Union[Tag]:
        while True:
            if self.pos == self.len - 1:
                break
            if self.is_start_tag():
                self.extract_start_tag()
            elif self.is_end_tag():
                self.extract_end_tag()
            elif self.is_in_tag():
                if self.is_tag_content():
                    self.consume_blank()
                    continue
                else:
                    self.extract_str_content()
            elif self.is_not_in_tag_str():
                self.extract_content_not_in_tag()
                                                           
        return self.root_tag




from byzerllm.utils.retrieval import ByzerRetrieval
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from byzerllm.utils.client import ByzerLLM
from byzerllm.utils import generate_str_md5
from byzerllm.utils.retrieval import ByzerRetrieval
import time
from byzerllm.utils.client import LLMHistoryItem,LLMRequest
from byzerllm.utils.retrieval import TableSettings,SearchQuery
import uuid
import json
from langchain_core.prompts import PromptTemplate
from langchain.text_splitter import RecursiveCharacterTextSplitter
try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

import jieba  

class SimpleRetrieval:
    def __init__(self,llm:ByzerLLM, retrieval: ByzerRetrieval,
                 chunk_collection: Optional[str] = "default",
                 retrieval_cluster:str="byzerai_store",
                 retrieval_db:str="byzerai_store",
                 max_output_length=10000):
        self.chunk_collection = chunk_collection
        self.retrieval_cluster = retrieval_cluster
        self.retrieval_db = retrieval_db
        self.max_output_length = max_output_length
        self.llm = llm
        self.retrieval = retrieval                
        if self.llm.default_emb_model_name is None:
            raise Exception(f'''
emb model does not exist. Try to use `llm.setup_default_emb_model_name` to set the default emb model name.
''')        

        # create the retrieval database/table if not exists
        if self.retrieval and not self.retrieval.check_table_exists(self.retrieval_cluster,self.retrieval_db,"text_content"):
           self.retrieval.create_table(self.retrieval_cluster,tableSettings=TableSettings(
                database=self.retrieval_db,
                table="text_content",schema='''st(
field(_id,string),
field(doc_id,string),
field(file_path,string),
field(owner,string),
field(content,string,analyze),
field(raw_content,string,no_index),
field(metadata,string,analyze),
field(json_data,string,no_index),
field(collection,string),
field(created_time,long,sort)
)''',
                location=f"/tmp/{self.retrieval_cluster}",num_shards=1 
           ))

           self.retrieval.create_table(self.retrieval_cluster,tableSettings=TableSettings(
                database=self.retrieval_db,
                table="text_content_chunk",schema='''st(
field(_id,string),
field(doc_id,string),
field(file_path,string),
field(owner,string),
field(chunk,string,analyze),
field(raw_chunk,string,no_index),
field(chunk_vector,array(float)),
field(metadata,string,analyze),
field(json_data,string,no_index),
field(chunk_collection,string),
field(created_time,long,sort)
)''',
                location=f"/tmp/{self.retrieval_cluster}",num_shards=1                
           )) 
            

    def save_doc(self,data:List[Dict[str,Any]],owner:Optional[str]=None,):

        if not self.retrieval:
            raise Exception("retrieval is not setup")
  
        owner = owner or "default"
    
        result = []
        for item in data:
            
            doc_id = item["doc_id"]
            json_data_obj = json.loads(item["json_data"])
            collection = item["collection"]
            _id = f"{collection}/{doc_id}"
                    
            file_path = json_data_obj.get("metadata",{}).get("file_path","")                
            result.append({"_id":_id,
            "doc_id":doc_id,     
            "file_path":file_path,          
            "json_data":item["json_data"],
            "metadata":self.search_tokenize(item["json_data"]), 
            "collection":collection,            
            "content":self.search_tokenize(item["content"]),
            "raw_content":item["content"],
            "owner":owner,          
            "created_time":int(time.time()*1000),
            })        
                    
        self.retrieval.build_from_dicts(self.retrieval_cluster,self.retrieval_db,"text_content",result)            

    
    def _owner_filter(self,owner:str):
        return {"field":"owner","value":owner}
            
    def search_content_chunks(self,
                              owner:str,
                              query_str:Optional[str]=None,
                              query_embedding:Optional[List[float]]=None,  
                              doc_ids:Optional[List[str]]=None,                            
                              limit:int=4,        
                              return_json:bool=True):   
        keyword = None
        fields = []
        vector = []
        vectorField = None
        filters = {"and":[{"field":"chunk_collection","value":self.chunk_collection}]}
    
        if query_str is not None:
            keyword = self.search_tokenize(query_str)
            fields = ["chunk"]

        if query_embedding is not None:
            vector = query_embedding
            vectorField = "chunk_vector"    
        
        
        if doc_ids:
            filters["and"].append({"or":[{"field":"doc_id","value":x} for x in doc_ids]})
        
        query = SearchQuery(self.retrieval_db,"text_content_chunk",
                                    filters=filters,
                                    keyword=keyword,fields=fields,
                                    vector=vector,vectorField=vectorField,
                                    limit=limit)
        
        docs = self.retrieval.search(self.retrieval_cluster,
                        [query])
    
        if return_json:
            context = json.dumps([{"content":x["raw_chunk"]} for x in docs],ensure_ascii=False,indent=4)    
            return context 
        else:
            return docs
        

    def search_content_by_filename(self,filename:str,collection:str): 
        filters = [{"field":"chunk_collection","value":collection}]
        docs = self.retrieval.search(self.retrieval_cluster,
                        [SearchQuery(self.retrieval_db,"text_content_chunk",
                                    filters={"and":filters},
                                    keyword=self.search_tokenize(filename),fields=["metadata"],
                                    vector=[],vectorField=None,
                                    limit=10000)])
        return docs 

    def search_content(self,q:str,owner:str,url:str,auth_tag:str=None,limit:int=4,return_json:bool=True): 
        filters = [self._owner_filter(owner)]
        
        if auth_tag:
            filters.append({"field":"auth_tag","value":self.search_tokenize(auth_tag)})
        
        if url:
            filters.append({"field":"url","value":url})    

        if q:
            keyword = self.search_tokenize(q)
            vector = self.emb(q)
            vectorField = "content_vector"
            fields = ["content"]
        else:
            keyword = None
            vector = []
            vectorField = None
            fields = []

        docs = self.retrieval.search(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content",
                                         filters={"and":filters},
                                        keyword=keyword,fields=fields,
                                        vector=vector,vectorField=vectorField,
                                        limit=limit)])

        if return_json:
            context = json.dumps([{"content":x["raw_content"]} for x in docs],ensure_ascii=False,indent=4)    
            return context 
        else:
            return docs 

    def get_chunk_by_id(self,chunk_id:str):
        filters = {"and":[{"field":"_id","value":chunk_id}]}        
        docs = self.retrieval.filter(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content_chunk",
                                         filters=filters,
                                        keyword=None,fields=[],
                                        vector=[],vectorField=None,
                                        limit=1)])
        return docs
    
    def get_chunks_by_docid(self,doc_id:str):
        docs = self.retrieval.filter(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content_chunk",
                                         filters={"and":[{"field":"doc_id","value":doc_id}]},
                                        keyword=None,fields=[],
                                        vector=[],vectorField=None,
                                        limit=100000)])
        return docs
    
    def delete_chunk_by_ids(self,chunk_ids:List[str]):        
        self.retrieval.delete_by_ids(self.retrieval_cluster,self.retrieval_db,"text_content_chunk",chunk_ids)
        self.retrieval.commit(self.retrieval_cluster,self.retrieval_db)    
    
    def save_chunks(self,chunks:List[Dict[str,Any]]):        
        text_content_chunks = []
        for chunk in chunks:
            chunk_id = chunk["chunk_id"]
            ref_doc_id = chunk["ref_doc_id"]
            owner = chunk.get("owner","default")
            chunk_content = chunk["chunk_content"]
            chunk_embedding = chunk["chunk_embedding"]
            metadata = chunk.get("metadata",{})
            _id = f"{self.chunk_collection}/{chunk_id}"
            file_path = metadata.get("file_path","")            
            
            text_content_chunks.append({"_id":_id,
                "doc_id":ref_doc_id or "",
                "file_path": file_path,
                "owner":owner or "default",                
                "chunk":self.search_tokenize(chunk_content),
                "raw_chunk":chunk_content,
                "chunk_vector":chunk_embedding,  
                "chunk_collection":self.chunk_collection,              
                "metadata":self.search_tokenize(json.dumps(metadata,ensure_ascii=False)),
                "json_data":json.dumps(metadata,ensure_ascii=False),
                "created_time":int(time.time()*1000),
                })                   
            
        self.retrieval.build_from_dicts(self.retrieval_cluster,
                                        self.retrieval_db,
                                        "text_content_chunk",text_content_chunks)           

            
        
    def get_doc(self,doc_id:str,collection:str):
        filters = {"and":[{"field":"_id","value":f'{collection}/{doc_id}'}]}        
        docs = self.retrieval.filter(self.retrieval_cluster,
                            [SearchQuery(self.retrieval_db,"text_content",
                                         filters=filters,
                                        keyword=None,fields=[],
                                        vector=[],vectorField=None,
                                        limit=1)])
        return docs[0] if docs else None
    
    @DeprecationWarning
    def delete_doc(self,doc_ids:List[str],collection:str):
        ids = []
        for doc_id in doc_ids:
            ids.append(f'{collection}/{doc_id}')
        self.retrieval.delete_by_ids(self.retrieval_cluster,self.retrieval_db,"text_content",[ids])

    def truncate_table(self):
        self.retrieval.truncate(self.retrieval_cluster,self.retrieval_db,"text_content")
        self.retrieval.truncate(self.retrieval_cluster,self.retrieval_db,"text_content_chunk")
        self.commit_doc()
        self.commit_chunk()

    def delete_doc_and_chunks_by_filename(self,collection:str,file_name:str):
        doc = self.get_doc(f"{collection}/ref_doc_info",file_name)
                
        if doc:
            node_ids = json.loads(doc["metadata"])["node_ids"]
            for node_id in node_ids:
                id = f"{collection}/data/{node_id}"
                self.retrieval.delete_by_ids(self.retrieval_cluster,self.retrieval_db,"text_content",[id])

                id = f"{collection}/metadata/{node_id}"
                self.retrieval.delete_by_ids(self.retrieval_cluster,self.retrieval_db,"text_content",[id])

                ## default/index/ can not be deleted
            
            ##  cleanup the chunk collection  
            self.retrieval.delete_by_filter(self.retrieval_cluster,self.retrieval_db,"text_content_chunk",
                                            {"file_path":file_name})                      


    def delete_from_doc_collection(self,collection:str):
        for suffix in ["index","data","ref_doc_info","metadata"]:
            self.retrieval.delete_by_filter(self.retrieval_cluster,self.retrieval_db,"text_content",
                                            {"collection":f"{collection}/{suffix}"})
        

    def delete_from_chunk_collection(self,collection:str):
        self.retrieval.delete_by_filter(self.retrieval_cluster,self.retrieval_db,"text_content_chunk",
                                        {"chunk_collection":collection})    

    def commit_doc(self):
        self.retrieval.commit(self.retrieval_cluster,self.retrieval_db,"text_content")    

    def commit_chunk(self):
        self.retrieval.commit(self.retrieval_cluster,self.retrieval_db,"text_content_chunk")    
                                  

    def emb(self,s:str):        
        return self.llm.emb(self.llm.default_emb_model_name,LLMRequest(instruction=s))[0].output


    def split_text_into_chunks(self,s:str):
        # self.llm.apply_sql_func(
        #     '''select llm_split(value,array(",","。","\n"),1600) as value ''',[{"value":content}],
        #     url=self.byzer_engine_url
        #     )["value"]
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=1600, chunk_overlap=200)
        split_docs = text_splitter.split_documents([Document(page_content=s)])         
        return [s.page_content for s in split_docs] 

    
    def search_tokenize(self,s:str):        
        seg_list = jieba.cut(s, cut_all=False)
        # return self.llm.apply_sql_func("select mkString(' ',parse(value)) as value",[
        # {"value":s}],url=self.byzer_engine_url)["value"]
        return " ".join(seg_list) 
        
    
               
    




import contextlib
import gc
import os
import re

from encodec import EncodecModel
import funcy
import logging
import numpy as np
from scipy.special import softmax
import torch
import torch.nn.functional as F
import tqdm
from transformers import BertTokenizer
from huggingface_hub import hf_hub_download

from .model import GPTConfig, GPT
from .model_fine import FineGPT, FineGPTConfig

if (
    torch.cuda.is_available() and
    hasattr(torch.cuda, "amp") and
    hasattr(torch.cuda.amp, "autocast") and
    hasattr(torch.cuda, "is_bf16_supported") and
    torch.cuda.is_bf16_supported()
):
    autocast = funcy.partial(torch.cuda.amp.autocast, dtype=torch.bfloat16)
else:
    @contextlib.contextmanager
    def autocast():
        yield

CONTEXT_WINDOW_SIZE = 1024

SEMANTIC_RATE_HZ = 49.9
SEMANTIC_VOCAB_SIZE = 10_000

CODEBOOK_SIZE = 1024
N_COARSE_CODEBOOKS = 2
N_FINE_CODEBOOKS = 8
COARSE_RATE_HZ = 75

SAMPLE_RATE = 24_000


SUPPORTED_LANGS = [
    ("English", "en"),
    ("German", "de"),
    ("Spanish", "es"),
    ("French", "fr"),
    ("Hindi", "hi"),
    ("Italian", "it"),
    ("Japanese", "ja"),
    ("Korean", "ko"),
    ("Polish", "pl"),
    ("Portuguese", "pt"),
    ("Russian", "ru"),
    ("Turkish", "tr"),
    ("Chinese", "zh"),
]

ALLOWED_PROMPTS = {"announcer"}
for _, lang in SUPPORTED_LANGS:
    for prefix in ("", f"v2{os.path.sep}"):
        for n in range(10):
            ALLOWED_PROMPTS.add(f"{prefix}{lang}_speaker_{n}")


logger = logging.getLogger(__name__)


CUR_PATH = os.path.dirname(os.path.abspath(__file__))


default_cache_dir = os.path.join(os.path.expanduser("~"), ".cache")
CACHE_DIR = os.path.join(os.getenv("XDG_CACHE_HOME", default_cache_dir), "suno", "bark_v0")


def _cast_bool_env_var(s):
    return s.lower() in ('true', '1', 't')


USE_SMALL_MODELS = _cast_bool_env_var(os.environ.get("SUNO_USE_SMALL_MODELS", "False"))
GLOBAL_ENABLE_MPS = _cast_bool_env_var(os.environ.get("SUNO_ENABLE_MPS", "False"))
OFFLOAD_CPU = _cast_bool_env_var(os.environ.get("SUNO_OFFLOAD_CPU", "False"))


REMOTE_MODEL_PATHS = {
    "text_small": {
        "repo_id": "suno/bark",
        "file_name": "text.pt",
    },
    "coarse_small": {
        "repo_id": "suno/bark",
        "file_name": "coarse.pt",
    },
    "fine_small": {
        "repo_id": "suno/bark",
        "file_name": "fine.pt",
    },
    "text": {
        "repo_id": "suno/bark",
        "file_name": "text_2.pt",
    },
    "coarse": {
        "repo_id": "suno/bark",
        "file_name": "coarse_2.pt",
    },
    "fine": {
        "repo_id": "suno/bark",
        "file_name": "fine_2.pt",
    },
}


if not hasattr(torch.nn.functional, 'scaled_dot_product_attention') and torch.cuda.is_available():
    logger.warning(
        "torch version does not support flash attention. You will get faster" +
        " inference speed by upgrade torch to newest nightly version."
    )


class InferenceContext:
    def __init__(self, benchmark=False):
        # we can't expect inputs to be the same length, so disable benchmarking by default
        self._chosen_cudnn_benchmark = benchmark
        self._cudnn_benchmark = None

    def __enter__(self):
        self._cudnn_benchmark = torch.backends.cudnn.benchmark
        torch.backends.cudnn.benchmark = self._chosen_cudnn_benchmark

    def __exit__(self, exc_type, exc_value, exc_traceback):
        torch.backends.cudnn.benchmark = self._cudnn_benchmark


if torch.cuda.is_available():
    torch.backends.cuda.matmul.allow_tf32 = True
    torch.backends.cudnn.allow_tf32 = True


@contextlib.contextmanager
def _inference_mode():
    with InferenceContext(), torch.inference_mode(), torch.no_grad(), autocast():
        yield


def _clear_cuda_cache():
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.synchronize()


TEXT_ENCODING_OFFSET = 10_048
SEMANTIC_PAD_TOKEN = 10_000
TEXT_PAD_TOKEN = 129_595
SEMANTIC_INFER_TOKEN = 129_599

COARSE_SEMANTIC_PAD_TOKEN = 12_048
COARSE_INFER_TOKEN = 12_050


class GenerateModel:
    # tokenizer_dir = "bert-base-multilingual-cased"
    def __init__(self, model_dir:str,tokenizer_dir:str):
        self.model_dir = model_dir if model_dir is not None else CACHE_DIR
        self.tokenizer_dir = tokenizer_dir
        self.models = {}
        self.models_devices = {}

    @staticmethod
    def _grab_best_device(use_gpu=True):
        if torch.cuda.device_count() > 0 and use_gpu:
            device = "cuda"
        elif torch.backends.mps.is_available() and use_gpu and GLOBAL_ENABLE_MPS:
            device = "mps"
        else:
            device = "cpu"
        return device


    def _get_ckpt_path(self,model_type, use_small=False):
        key = model_type
        if use_small or USE_SMALL_MODELS:
            key += "_small"

        return os.path.join(self.model_dir, REMOTE_MODEL_PATHS[key]["file_name"])


    def _download(self,from_hf_path, file_name):
        os.makedirs(self.model_dir, exist_ok=True)
        hf_hub_download(repo_id=from_hf_path, filename=file_name, local_dir=CACHE_DIR)
    
    def clean_models(self,model_key=None):        
        model_keys = [model_key] if model_key is not None else self.models.keys()
        for k in model_keys:
            if k in self.models:
                del self.models[k]
        _clear_cuda_cache()
        gc.collect()


    def _load_model(self,ckpt_path, device, use_small=False, model_type="text"):
        if model_type == "text":
            ConfigClass = GPTConfig
            ModelClass = GPT
        elif model_type == "coarse":
            ConfigClass = GPTConfig
            ModelClass = GPT
        elif model_type == "fine":
            ConfigClass = FineGPTConfig
            ModelClass = FineGPT
        else:
            raise NotImplementedError()
        model_key = f"{model_type}_small" if use_small or USE_SMALL_MODELS else model_type
        model_info = REMOTE_MODEL_PATHS[model_key]
        if not os.path.exists(ckpt_path):
            logger.info(f"{model_type} model not found, downloading into `{self.model_dir}`.")
            self._download(model_info["repo_id"], model_info["file_name"])
        
        print(f"loading {model_type} model from `{ckpt_path}`")    

        checkpoint = torch.load(ckpt_path, map_location=device)
        # this is a hack
        model_args = checkpoint["model_args"]
        if "input_vocab_size" not in model_args:
            model_args["input_vocab_size"] = model_args["vocab_size"]
            model_args["output_vocab_size"] = model_args["vocab_size"]
            del model_args["vocab_size"]
        gptconf = ConfigClass(**checkpoint["model_args"])
        model = ModelClass(gptconf)
        state_dict = checkpoint["model"]
        # fixup checkpoint
        unwanted_prefix = "_orig_mod."
        for k, v in list(state_dict.items()):
            if k.startswith(unwanted_prefix):
                state_dict[k[len(unwanted_prefix) :]] = state_dict.pop(k)
        extra_keys = set(state_dict.keys()) - set(model.state_dict().keys())
        extra_keys = set([k for k in extra_keys if not k.endswith(".attn.bias")])
        missing_keys = set(model.state_dict().keys()) - set(state_dict.keys())
        missing_keys = set([k for k in missing_keys if not k.endswith(".attn.bias")])
        if len(extra_keys) != 0:
            raise ValueError(f"extra keys found: {extra_keys}")
        if len(missing_keys) != 0:
            raise ValueError(f"missing keys: {missing_keys}")
        model.load_state_dict(state_dict, strict=False)
        n_params = model.get_num_params()
        val_loss = checkpoint["best_val_loss"].item()
        logger.info(f"model loaded: {round(n_params/1e6,1)}M params, {round(val_loss,3)} loss")
        model.eval()
        model.to(device)
        del checkpoint, state_dict
        _clear_cuda_cache()
        if model_type == "text":
            tokenizer = BertTokenizer.from_pretrained(self.tokenizer_dir)
            return {
                "model": model,
                "tokenizer": tokenizer,
            }
        return model


    def _load_codec_model(self,device):
        model = EncodecModel.encodec_model_24khz()
        model.set_target_bandwidth(6.0)
        model.eval()
        model.to(device)
        _clear_cuda_cache()
        return model


    def load_model(self,use_gpu=True, use_small=False, force_reload=False, model_type="text"):
        ## _load_model_f = funcy.partial(self._load_model, model_type=model_type, use_small=use_small)
        if model_type not in ("text", "coarse", "fine"):
            raise NotImplementedError()        
        device = GenerateModel._grab_best_device(use_gpu=use_gpu)
        model_key = f"{model_type}"
        if OFFLOAD_CPU:
            self.models_devices[model_key] = device
            device = "cpu"
        if model_key not in self.models or force_reload:
            ckpt_path = self._get_ckpt_path(model_type, use_small=use_small)
            self.clean_models(model_key=model_key)
            model = self._load_model(ckpt_path = ckpt_path, device = device,use_small = use_small, model_type=model_type)
            self.models[model_key] = model
        if model_type == "text":
            self.models[model_key]["model"].to(device)
        else:
            self.models[model_key].to(device)
        return self.models[model_key]


    def load_codec_model(self,use_gpu=True, force_reload=False):        
        device = GenerateModel._grab_best_device(use_gpu=use_gpu)
        if device == "mps":
            # encodec doesn't support mps
            device = "cpu"
        model_key = "codec"
        if OFFLOAD_CPU:
            self.models_devices[model_key] = device
            device = "cpu"
        if model_key not in self.models or force_reload:
            self.clean_models(model_key=model_key)
            model = self._load_codec_model(device)
            self.models[model_key] = model
        self.models[model_key].to(device)
        return self.models[model_key]

    
    def preload_models(self,
        text_use_gpu=True,
        text_use_small=False,
        coarse_use_gpu=True,
        coarse_use_small=False,
        fine_use_gpu=True,
        fine_use_small=False,
        codec_use_gpu=True,
        force_reload=False,
    ):
        """Load all the necessary models for the pipeline."""
        if GenerateModel._grab_best_device() == "cpu" and (
            text_use_gpu or coarse_use_gpu or fine_use_gpu or codec_use_gpu
        ):
            logger.warning("No GPU being used. Careful, inference might be very slow!")
        _ = self.load_model(
            model_type="text", use_gpu=text_use_gpu, use_small=text_use_small, force_reload=force_reload
        )
        _ = self.load_model(
            model_type="coarse",
            use_gpu=coarse_use_gpu,
            use_small=coarse_use_small,
            force_reload=force_reload,
        )
        _ = self.load_model(
            model_type="fine", use_gpu=fine_use_gpu, use_small=fine_use_small, force_reload=force_reload
        )
        _ = self.load_codec_model(use_gpu=codec_use_gpu, force_reload=force_reload)


    @staticmethod
    def _tokenize(tokenizer, text):
        return tokenizer.encode(text, add_special_tokens=False)

    @staticmethod
    def _detokenize(tokenizer, enc_text):
        return tokenizer.decode(enc_text)

    @staticmethod
    def _normalize_whitespace(text):
        return re.sub(r"\s+", " ", text).strip()

    @staticmethod 
    def _load_history_prompt(history_prompt_input):
        if isinstance(history_prompt_input, str) and history_prompt_input.endswith(".npz"):
            history_prompt = np.load(history_prompt_input)
        elif isinstance(history_prompt_input, str):
            # make sure this works on non-ubuntu
            history_prompt_input = os.path.join(*history_prompt_input.split("/"))
            if history_prompt_input not in ALLOWED_PROMPTS:
                raise ValueError("history prompt not found")
            history_prompt = np.load(
                os.path.join(CUR_PATH, "assets", "prompts", f"{history_prompt_input}.npz")
            )
        elif isinstance(history_prompt_input, dict):
            assert("semantic_prompt" in history_prompt_input)
            assert("coarse_prompt" in history_prompt_input)
            assert("fine_prompt" in history_prompt_input)
            history_prompt = history_prompt_input
        else:
            raise ValueError("history prompt format unrecognized")
        return history_prompt
    
    def generate_text_semantic(
        self,
        text,
        history_prompt=None,
        temp=0.7,
        top_k=None,
        top_p=None,
        silent=False,
        min_eos_p=0.2,
        max_gen_duration_s=None,
        allow_early_stop=True,
        use_kv_caching=False,
    ):
        """Generate semantic tokens from text."""
        assert isinstance(text, str)
        text = GenerateModel._normalize_whitespace(text)
        assert len(text.strip()) > 0
        if history_prompt is not None:
            history_prompt = GenerateModel._load_history_prompt(history_prompt)
            semantic_history = history_prompt["semantic_prompt"]
            assert (
                isinstance(semantic_history, np.ndarray)
                and len(semantic_history.shape) == 1
                and len(semantic_history) > 0
                and semantic_history.min() >= 0
                and semantic_history.max() <= SEMANTIC_VOCAB_SIZE - 1
            )
        else:
            semantic_history = None
        # load models if not yet exist        
        if "text" not in self.models:
            self.preload_models()
        model_container = self.models["text"]
        model = model_container["model"]
        tokenizer = model_container["tokenizer"]
        encoded_text = np.array(GenerateModel._tokenize(tokenizer, text)) + TEXT_ENCODING_OFFSET
        if OFFLOAD_CPU:
            model.to(self.models_devices["text"])
        device = next(model.parameters()).device
        if len(encoded_text) > 256:
            p = round((len(encoded_text) - 256) / len(encoded_text) * 100, 1)
            logger.warning(f"warning, text too long, lopping of last {p}%")
            encoded_text = encoded_text[:256]
        encoded_text = np.pad(
            encoded_text,
            (0, 256 - len(encoded_text)),
            constant_values=TEXT_PAD_TOKEN,
            mode="constant",
        )
        if semantic_history is not None:
            semantic_history = semantic_history.astype(np.int64)
            # lop off if history is too long, pad if needed
            semantic_history = semantic_history[-256:]
            semantic_history = np.pad(
                semantic_history,
                (0, 256 - len(semantic_history)),
                constant_values=SEMANTIC_PAD_TOKEN,
                mode="constant",
            )
        else:
            semantic_history = np.array([SEMANTIC_PAD_TOKEN] * 256)
        x = torch.from_numpy(
            np.hstack([
                encoded_text, semantic_history, np.array([SEMANTIC_INFER_TOKEN])
            ]).astype(np.int64)
        )[None]
        assert x.shape[1] == 256 + 256 + 1
        with _inference_mode():
            x = x.to(device)
            n_tot_steps = 768
            # custom tqdm updates since we don't know when eos will occur
            pbar = tqdm.tqdm(disable=silent, total=100)
            pbar_state = 0
            tot_generated_duration_s = 0
            kv_cache = None
            for n in range(n_tot_steps):
                if use_kv_caching and kv_cache is not None:
                    x_input = x[:, [-1]]
                else:
                    x_input = x
                logits, kv_cache = model(
                    x_input, merge_context=True, use_cache=use_kv_caching, past_kv=kv_cache
                )
                relevant_logits = logits[0, 0, :SEMANTIC_VOCAB_SIZE]
                if allow_early_stop:
                    relevant_logits = torch.hstack(
                        (relevant_logits, logits[0, 0, [SEMANTIC_PAD_TOKEN]])  # eos
                    )
                if top_p is not None:
                    # faster to convert to numpy
                    logits_device = relevant_logits.device
                    logits_dtype = relevant_logits.type()
                    relevant_logits = relevant_logits.detach().cpu().type(torch.float32).numpy()
                    sorted_indices = np.argsort(relevant_logits)[::-1]
                    sorted_logits = relevant_logits[sorted_indices]
                    cumulative_probs = np.cumsum(softmax(sorted_logits))
                    sorted_indices_to_remove = cumulative_probs > top_p
                    sorted_indices_to_remove[1:] = sorted_indices_to_remove[:-1].copy()
                    sorted_indices_to_remove[0] = False
                    relevant_logits[sorted_indices[sorted_indices_to_remove]] = -np.inf
                    relevant_logits = torch.from_numpy(relevant_logits)
                    relevant_logits = relevant_logits.to(logits_device).type(logits_dtype)
                if top_k is not None:
                    v, _ = torch.topk(relevant_logits, min(top_k, relevant_logits.size(-1)))
                    relevant_logits[relevant_logits < v[-1]] = -float("Inf")
                probs = F.softmax(relevant_logits / temp, dim=-1)
                # multinomial bugged on mps: shuttle to cpu if necessary
                inf_device = probs.device
                if probs.device.type == "mps":
                    probs = probs.to("cpu")
                item_next = torch.multinomial(probs, num_samples=1)
                probs = probs.to(inf_device)
                item_next = item_next.to(inf_device)
                if allow_early_stop and (
                    item_next == SEMANTIC_VOCAB_SIZE
                    or (min_eos_p is not None and probs[-1] >= min_eos_p)
                ):
                    # eos found, so break
                    pbar.update(100 - pbar_state)
                    break
                x = torch.cat((x, item_next[None]), dim=1)
                tot_generated_duration_s += 1 / SEMANTIC_RATE_HZ
                if max_gen_duration_s is not None and tot_generated_duration_s > max_gen_duration_s:
                    pbar.update(100 - pbar_state)
                    break
                if n == n_tot_steps - 1:
                    pbar.update(100 - pbar_state)
                    break
                del logits, relevant_logits, probs, item_next
                req_pbar_state = np.min([100, int(round(100 * n / n_tot_steps))])
                if req_pbar_state > pbar_state:
                    pbar.update(req_pbar_state - pbar_state)
                pbar_state = req_pbar_state
            pbar.close()
            out = x.detach().cpu().numpy().squeeze()[256 + 256 + 1 :]
        if OFFLOAD_CPU:
            model.to("cpu")
        assert all(0 <= out) and all(out < SEMANTIC_VOCAB_SIZE)
        _clear_cuda_cache()
        return out    

    @staticmethod
    def _flatten_codebooks(arr, offset_size=CODEBOOK_SIZE):
        assert len(arr.shape) == 2
        arr = arr.copy()
        if offset_size is not None:
            for n in range(1, arr.shape[0]):
                arr[n, :] += offset_size * n
        flat_arr = arr.ravel("F")
        return flat_arr    


    def generate_coarse(self,
        x_semantic,
        history_prompt=None,
        temp=0.7,
        top_k=None,
        top_p=None,
        silent=False,
        max_coarse_history=630,  # min 60 (faster), max 630 (more context)
        sliding_window_len=60,
        use_kv_caching=False,
    ):
        """Generate coarse audio codes from semantic tokens."""
        assert (
            isinstance(x_semantic, np.ndarray)
            and len(x_semantic.shape) == 1
            and len(x_semantic) > 0
            and x_semantic.min() >= 0
            and x_semantic.max() <= SEMANTIC_VOCAB_SIZE - 1
        )
        assert 60 <= max_coarse_history <= 630
        assert max_coarse_history + sliding_window_len <= 1024 - 256
        semantic_to_coarse_ratio = COARSE_RATE_HZ / SEMANTIC_RATE_HZ * N_COARSE_CODEBOOKS
        max_semantic_history = int(np.floor(max_coarse_history / semantic_to_coarse_ratio))
        if history_prompt is not None:
            history_prompt = GenerateModel._load_history_prompt(history_prompt)
            x_semantic_history = history_prompt["semantic_prompt"]
            x_coarse_history = history_prompt["coarse_prompt"]
            assert (
                isinstance(x_semantic_history, np.ndarray)
                and len(x_semantic_history.shape) == 1
                and len(x_semantic_history) > 0
                and x_semantic_history.min() >= 0
                and x_semantic_history.max() <= SEMANTIC_VOCAB_SIZE - 1
                and isinstance(x_coarse_history, np.ndarray)
                and len(x_coarse_history.shape) == 2
                and x_coarse_history.shape[0] == N_COARSE_CODEBOOKS
                and x_coarse_history.shape[-1] >= 0
                and x_coarse_history.min() >= 0
                and x_coarse_history.max() <= CODEBOOK_SIZE - 1
                and (
                    round(x_coarse_history.shape[-1] / len(x_semantic_history), 1)
                    == round(semantic_to_coarse_ratio / N_COARSE_CODEBOOKS, 1)
                )
            )
            x_coarse_history = GenerateModel._flatten_codebooks(x_coarse_history) + SEMANTIC_VOCAB_SIZE
            # trim histories correctly
            n_semantic_hist_provided = np.min(
                [
                    max_semantic_history,
                    len(x_semantic_history) - len(x_semantic_history) % 2,
                    int(np.floor(len(x_coarse_history) / semantic_to_coarse_ratio)),
                ]
            )
            n_coarse_hist_provided = int(round(n_semantic_hist_provided * semantic_to_coarse_ratio))
            x_semantic_history = x_semantic_history[-n_semantic_hist_provided:].astype(np.int32)
            x_coarse_history = x_coarse_history[-n_coarse_hist_provided:].astype(np.int32)
            # TODO: bit of a hack for time alignment (sounds better)
            x_coarse_history = x_coarse_history[:-2]
        else:
            x_semantic_history = np.array([], dtype=np.int32)
            x_coarse_history = np.array([], dtype=np.int32)
        # load models if not yet exist        
        if "coarse" not in self.models:
            self.preload_models()
        model = self.models["coarse"]
        if OFFLOAD_CPU:
            model.to(self.models_devices["coarse"])
        device = next(model.parameters()).device
        # start loop
        n_steps = int(
            round(
                np.floor(len(x_semantic) * semantic_to_coarse_ratio / N_COARSE_CODEBOOKS)
                * N_COARSE_CODEBOOKS
            )
        )
        assert n_steps > 0 and n_steps % N_COARSE_CODEBOOKS == 0
        x_semantic = np.hstack([x_semantic_history, x_semantic]).astype(np.int32)
        x_coarse = x_coarse_history.astype(np.int32)
        base_semantic_idx = len(x_semantic_history)
        with _inference_mode():
            x_semantic_in = torch.from_numpy(x_semantic)[None].to(device)
            x_coarse_in = torch.from_numpy(x_coarse)[None].to(device)
            n_window_steps = int(np.ceil(n_steps / sliding_window_len))
            n_step = 0
            for _ in tqdm.tqdm(range(n_window_steps), total=n_window_steps, disable=silent):
                semantic_idx = base_semantic_idx + int(round(n_step / semantic_to_coarse_ratio))
                # pad from right side
                x_in = x_semantic_in[:, np.max([0, semantic_idx - max_semantic_history]) :]
                x_in = x_in[:, :256]
                x_in = F.pad(
                    x_in,
                    (0, 256 - x_in.shape[-1]),
                    "constant",
                    COARSE_SEMANTIC_PAD_TOKEN,
                )
                x_in = torch.hstack(
                    [
                        x_in,
                        torch.tensor([COARSE_INFER_TOKEN])[None].to(device),
                        x_coarse_in[:, -max_coarse_history:],
                    ]
                )
                kv_cache = None
                for _ in range(sliding_window_len):
                    if n_step >= n_steps:
                        continue
                    is_major_step = n_step % N_COARSE_CODEBOOKS == 0

                    if use_kv_caching and kv_cache is not None:
                        x_input = x_in[:, [-1]]
                    else:
                        x_input = x_in

                    logits, kv_cache = model(x_input, use_cache=use_kv_caching, past_kv=kv_cache)
                    logit_start_idx = (
                        SEMANTIC_VOCAB_SIZE + (1 - int(is_major_step)) * CODEBOOK_SIZE
                    )
                    logit_end_idx = (
                        SEMANTIC_VOCAB_SIZE + (2 - int(is_major_step)) * CODEBOOK_SIZE
                    )
                    relevant_logits = logits[0, 0, logit_start_idx:logit_end_idx]
                    if top_p is not None:
                        # faster to convert to numpy
                        logits_device = relevant_logits.device
                        logits_dtype = relevant_logits.type()
                        relevant_logits = relevant_logits.detach().cpu().type(torch.float32).numpy()
                        sorted_indices = np.argsort(relevant_logits)[::-1]
                        sorted_logits = relevant_logits[sorted_indices]
                        cumulative_probs = np.cumsum(softmax(sorted_logits))
                        sorted_indices_to_remove = cumulative_probs > top_p
                        sorted_indices_to_remove[1:] = sorted_indices_to_remove[:-1].copy()
                        sorted_indices_to_remove[0] = False
                        relevant_logits[sorted_indices[sorted_indices_to_remove]] = -np.inf
                        relevant_logits = torch.from_numpy(relevant_logits)
                        relevant_logits = relevant_logits.to(logits_device).type(logits_dtype)
                    if top_k is not None:
                        v, _ = torch.topk(relevant_logits, min(top_k, relevant_logits.size(-1)))
                        relevant_logits[relevant_logits < v[-1]] = -float("Inf")
                    probs = F.softmax(relevant_logits / temp, dim=-1)
                    # multinomial bugged on mps: shuttle to cpu if necessary
                    inf_device = probs.device
                    if probs.device.type == "mps":
                        probs = probs.to("cpu")
                    item_next = torch.multinomial(probs, num_samples=1)
                    probs = probs.to(inf_device)
                    item_next = item_next.to(inf_device)
                    item_next += logit_start_idx
                    x_coarse_in = torch.cat((x_coarse_in, item_next[None]), dim=1)
                    x_in = torch.cat((x_in, item_next[None]), dim=1)
                    del logits, relevant_logits, probs, item_next
                    n_step += 1
                del x_in
            del x_semantic_in
        if OFFLOAD_CPU:
            model.to("cpu")
        gen_coarse_arr = x_coarse_in.detach().cpu().numpy().squeeze()[len(x_coarse_history) :]
        del x_coarse_in
        assert len(gen_coarse_arr) == n_steps
        gen_coarse_audio_arr = gen_coarse_arr.reshape(-1, N_COARSE_CODEBOOKS).T - SEMANTIC_VOCAB_SIZE
        for n in range(1, N_COARSE_CODEBOOKS):
            gen_coarse_audio_arr[n, :] -= n * CODEBOOK_SIZE
        _clear_cuda_cache()
        return gen_coarse_audio_arr


    def generate_fine(self,
        x_coarse_gen,
        history_prompt=None,
        temp=0.5,
        silent=True,
    ):
        """Generate full audio codes from coarse audio codes."""
        assert (
            isinstance(x_coarse_gen, np.ndarray)
            and len(x_coarse_gen.shape) == 2
            and 1 <= x_coarse_gen.shape[0] <= N_FINE_CODEBOOKS - 1
            and x_coarse_gen.shape[1] > 0
            and x_coarse_gen.min() >= 0
            and x_coarse_gen.max() <= CODEBOOK_SIZE - 1
        )
        if history_prompt is not None:
            history_prompt = GenerateModel._load_history_prompt(history_prompt)
            x_fine_history = history_prompt["fine_prompt"]
            assert (
                isinstance(x_fine_history, np.ndarray)
                and len(x_fine_history.shape) == 2
                and x_fine_history.shape[0] == N_FINE_CODEBOOKS
                and x_fine_history.shape[1] >= 0
                and x_fine_history.min() >= 0
                and x_fine_history.max() <= CODEBOOK_SIZE - 1
            )
        else:
            x_fine_history = None
        n_coarse = x_coarse_gen.shape[0]
        # load models if not yet exist        
        if "fine" not in self.models:
            self.preload_models()
        model = self.models["fine"]
        if OFFLOAD_CPU:
            model.to(self.models_devices["fine"])
        device = next(model.parameters()).device
        # make input arr
        in_arr = np.vstack(
            [
                x_coarse_gen,
                np.zeros((N_FINE_CODEBOOKS - n_coarse, x_coarse_gen.shape[1]))
                + CODEBOOK_SIZE,  # padding
            ]
        ).astype(np.int32)
        # prepend history if available (max 512)
        if x_fine_history is not None:
            x_fine_history = x_fine_history.astype(np.int32)
            in_arr = np.hstack(
                [
                    x_fine_history[:, -512:].astype(np.int32),
                    in_arr,
                ]
            )
            n_history = x_fine_history[:, -512:].shape[1]
        else:
            n_history = 0
        n_remove_from_end = 0
        # need to pad if too short (since non-causal model)
        if in_arr.shape[1] < 1024:
            n_remove_from_end = 1024 - in_arr.shape[1]
            in_arr = np.hstack(
                [
                    in_arr,
                    np.zeros((N_FINE_CODEBOOKS, n_remove_from_end), dtype=np.int32) + CODEBOOK_SIZE,
                ]
            )
        # we can be lazy about fractional loop and just keep overwriting codebooks
        n_loops = np.max([0, int(np.ceil((x_coarse_gen.shape[1] - (1024 - n_history)) / 512))]) + 1
        with _inference_mode():
            in_arr = torch.tensor(in_arr.T).to(device)
            for n in tqdm.tqdm(range(n_loops), disable=silent):
                start_idx = np.min([n * 512, in_arr.shape[0] - 1024])
                start_fill_idx = np.min([n_history + n * 512, in_arr.shape[0] - 512])
                rel_start_fill_idx = start_fill_idx - start_idx
                in_buffer = in_arr[start_idx : start_idx + 1024, :][None]
                for nn in range(n_coarse, N_FINE_CODEBOOKS):
                    logits = model(nn, in_buffer)
                    if temp is None:
                        relevant_logits = logits[0, rel_start_fill_idx:, :CODEBOOK_SIZE]
                        codebook_preds = torch.argmax(relevant_logits, -1)
                    else:
                        relevant_logits = logits[0, :, :CODEBOOK_SIZE] / temp
                        probs = F.softmax(relevant_logits, dim=-1)
                        # multinomial bugged on mps: shuttle to cpu if necessary
                        inf_device = probs.device
                        if probs.device.type == "mps":
                            probs = probs.to("cpu")
                        codebook_preds = torch.hstack(
                            [
                                torch.multinomial(probs[nnn], num_samples=1).to(inf_device)
                                for nnn in range(rel_start_fill_idx, 1024)
                            ]
                        )
                    in_buffer[0, rel_start_fill_idx:, nn] = codebook_preds
                    del logits, codebook_preds
                # transfer over info into model_in and convert to numpy
                for nn in range(n_coarse, N_FINE_CODEBOOKS):
                    in_arr[
                        start_fill_idx : start_fill_idx + (1024 - rel_start_fill_idx), nn
                    ] = in_buffer[0, rel_start_fill_idx:, nn]
                del in_buffer
            gen_fine_arr = in_arr.detach().cpu().numpy().squeeze().T
            del in_arr
        if OFFLOAD_CPU:
            model.to("cpu")
        gen_fine_arr = gen_fine_arr[:, n_history:]
        if n_remove_from_end > 0:
            gen_fine_arr = gen_fine_arr[:, :-n_remove_from_end]
        assert gen_fine_arr.shape[-1] == x_coarse_gen.shape[-1]
        _clear_cuda_cache()
        return gen_fine_arr


    def codec_decode(self,fine_tokens):
        """Turn quantized audio codes into audio array using encodec."""
        
        if "codec" not in self.models:
            self.preload_models()
        model = self.models["codec"]
        if OFFLOAD_CPU:
            model.to(self.models_devices["codec"])
        device = next(model.parameters()).device
        arr = torch.from_numpy(fine_tokens)[None]
        arr = arr.to(device)
        arr = arr.transpose(0, 1)
        emb = model.quantizer.decode(arr)
        out = model.decoder(emb)
        audio_arr = out.detach().cpu().numpy().squeeze()
        del arr, emb, out
        if OFFLOAD_CPU:
            model.to("cpu")
        return audio_arr



import copy
from typing import *

import torch
from diffusers import (
    AutoencoderKL,
    DDPMScheduler,
    EulerAncestralDiscreteScheduler,
    KDPM2AncestralDiscreteScheduler,
    UNet2DConditionModel,
)
from tqdm import tqdm
from transformers import CLIPTextModel, CLIPTokenizer

from byzerllm.stable_diffusion.diffusion.upscalers.samplers import (
    EulerAncestralSampler,
    KDPM2AncestralSampler,
)


class Multidiffusion:
    def __init__(
        self,
        pipe,
    ):
        self.vae: AutoencoderKL = pipe.text_encoder
        self.text_encoder: CLIPTextModel = pipe.text_encoder
        self.tokenizer: CLIPTokenizer = pipe.tokenizer
        self.unet: UNet2DConditionModel = pipe.unet
        self.scheduler: DDPMScheduler = pipe.scheduler
        self.ancestral = False

    def hijack_ancestral_scheduler(self) -> bool:
        if isinstance(self.scheduler, EulerAncestralDiscreteScheduler):
            config = copy.deepcopy(self.scheduler.__dict__)
            self.scheduler = EulerAncestralSampler.from_config(self.scheduler.config)
            self.scheduler.__dict__.update(config)
            return True
        elif isinstance(self.scheduler, KDPM2AncestralDiscreteScheduler):
            config = copy.deepcopy(self.scheduler.__dict__)
            self.scheduler = KDPM2AncestralSampler.from_config(self.scheduler.config)
            self.scheduler.__dict__.update(config)
            return True
        else:
            return False

    @classmethod
    def get_views(cls, panorama_height, panorama_width, window_size=64, stride=8):
        # Here, we define the mappings F_i (see Eq. 7 in the MultiDiffusion paper https://arxiv.org/abs/2302.08113)
        panorama_height /= 8
        panorama_width /= 8
        num_blocks_height = (
            (panorama_height - window_size) // stride + 1
            if panorama_height > window_size
            else 1
        )
        num_blocks_width = (
            (panorama_width - window_size) // stride + 1
            if panorama_width > window_size
            else 1
        )
        total_num_blocks = int(num_blocks_height * num_blocks_width)
        views = []
        for i in range(total_num_blocks):
            h_start = int((i // num_blocks_width) * stride)
            h_end = h_start + window_size
            w_start = int((i % num_blocks_width) * stride)
            w_end = w_start + window_size
            views.append((h_start, h_end, w_start, w_end))
        return views

    def align_unet_inputs(
        self,
        latent_model_input: torch.Tensor,
        prompt_embeds: torch.Tensor,
        views_batch_size: int,
        real_batch_size: int,
    ):
        prompt_embeds_align = torch.cat([prompt_embeds] * real_batch_size)
        latent_align = latent_model_input
        return latent_align, prompt_embeds_align

    def views_denoise_latent(
        self,
        views: list,
        latents: torch.Tensor,
        timesteps: torch.Tensor,
        num_inference_steps: int,
        guidance_scale: float,
        do_classifier_free_guidance: bool,
        prompt_embeds: torch.Tensor,
        extra_step_kwargs: Dict[str, Any],
        callback: Optional[Callable],
        callback_steps: int,
        cross_attention_kwargs: Dict[str, Any],
        views_batch_size: int = 1,
    ):
        # hijack ancestral schedulers
        self.ancestral = self.hijack_ancestral_scheduler()
        # 6. Define panorama grid and initialize views for synthesis.
        views_batch = [
            views[i : i + views_batch_size]
            for i in range(0, len(views), views_batch_size)
        ]
        views_scheduler_status = [copy.deepcopy(self.scheduler.__dict__)] * len(
            views_batch
        )
        count = torch.zeros_like(latents)
        value = torch.zeros_like(latents)
        num_warmup_steps = len(timesteps) - num_inference_steps * self.scheduler.order
        # 7. multidiffusion denoise loop
        with tqdm(total=num_inference_steps) as progress_bar:
            for step, timestep in enumerate(timesteps):
                count.zero_()
                value.zero_()
                noise = torch.randn_like(latents)
                for j, batch_view in enumerate(views_batch):
                    vb_size = len(batch_view)
                    # get the latents corresponding to the current view coordinates
                    latents_for_view = torch.cat(
                        [
                            latents[:, :, h_start:h_end, w_start:w_end]
                            for h_start, h_end, w_start, w_end in batch_view
                        ]
                    )
                    self.scheduler.__dict__.update(views_scheduler_status[j])

                    # expand the latents if we are doing classifier free guidance
                    latent_model_input = (
                        latents_for_view.repeat_interleave(2, dim=0)
                        if do_classifier_free_guidance
                        else latents_for_view
                    )
                    latent_model_input = self.scheduler.scale_model_input(
                        latent_model_input, timestep
                    )

                    # align unet inputs for batch
                    latent_model_input, prompt_embeds_input = self.align_unet_inputs(
                        latent_model_input=latent_model_input,
                        prompt_embeds=prompt_embeds,
                        views_batch_size=views_batch_size,
                        real_batch_size=vb_size,
                    )

                    # predict the noise residual
                    noise_pred = self.unet(
                        latent_model_input,
                        timestep,
                        encoder_hidden_states=prompt_embeds_input,
                        cross_attention_kwargs=cross_attention_kwargs,
                    ).sample

                    # perform guidance
                    if do_classifier_free_guidance:
                        noise_pred_uncond, noise_pred_text = (
                            noise_pred[::2],
                            noise_pred[1::2],
                        )
                        noise_pred = noise_pred_uncond + guidance_scale * (
                            noise_pred_text - noise_pred_uncond
                        )
                        noise_pred = noise_pred[:vb_size]

                    # compute the previous noisy sample x_t -> x_t-1
                    scheduler_output = self.scheduler.step(
                        model_output=noise_pred,
                        timestep=timestep,
                        sample=latents_for_view,
                        **extra_step_kwargs,
                    )
                    latents_denoised_batch = scheduler_output.prev_sample
                    sigma_up = scheduler_output.sigma_up if self.ancestral else None

                    views_scheduler_status[j] = copy.deepcopy(self.scheduler.__dict__)

                    # extract value from batch
                    for latents_view_denoised, (h_start, h_end, w_start, w_end) in zip(
                        latents_denoised_batch.chunk(vb_size), batch_view
                    ):
                        value[
                            :, :, h_start:h_end, w_start:w_end
                        ] += latents_view_denoised
                        count[:, :, h_start:h_end, w_start:w_end] += 1

                # take the MultiDiffusion step. Eq. 5 in MultiDiffusion paper: https://arxiv.org/abs/2302.08113
                # add noise for ancestral sampler
                latents = (
                    torch.where(count > 0, value / count, value) + noise * sigma_up
                    if sigma_up
                    else torch.where(count > 0, value / count, value)
                )

                # call the callback, if provided
                if step == len(timesteps) - 1 or (
                    (step + 1) > num_warmup_steps
                    and (step + 1) % self.scheduler.order == 0
                ):
                    progress_bar.update()
                    if callback is not None and step % callback_steps == 0:
                        callback(step, timestep, latents)

        return 1 / 0.18215 * latents



from langchain.embeddings.base import Embeddings
from typing import List, Union
import torch
import torch.nn.functional as F
from transformers import pipeline

class ByzerSentenceTransformerEmbeddings(Embeddings):
    
    def __init__(self, model,tokenizer, device="auto"):         
        if model:
            self.model = model
        if tokenizer:
            self.model = tokenizer            
        
        if device == "auto":
            self.device = (
                torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
            )
        else:
            self.device = device
        
        
    def _encode(self,texts: List[str],extract_params={}):        
        embeddings = [emb.tolist() for emb in self.model.encode(texts,**extract_params)]
        return embeddings
        
    def embed_documents(self, texts: List[str],extract_params={}) -> List[List[float]]:        
        embeddings = self._encode(texts,extract_params)
        return embeddings

    def embed_query(self, text: str,extract_params={}) -> List[float]:    
        embedding = self._encode([text],extract_params)
        return embedding[0]
        

class ByzerLLMEmbeddings(Embeddings):
    def __init__(self, model,tokenizer, device="auto", use_feature_extraction=False):         
        self.model = model
        self.tokenizer = tokenizer
        
        if device == "auto":
            self.device = (
                torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
            )
        else:
            self.device = device

        self.pipeline = None
        if use_feature_extraction:
            self.pipeline = pipeline("feature-extraction", model = model, tokenizer = tokenizer,device=0)
        
    def _encode(self,texts: List[str],extract_params={}):
        if self.pipeline:
            return [self.pipeline(text)[0][-1] for text in texts]
        else:
            _, embeddings = self.get_embedding_with_token_count(texts)
            embeddings = embeddings.detach().cpu()
            embeddings = embeddings.numpy()
            embeddings = [emb.tolist() for emb in embeddings]
            return embeddings
        
    def embed_documents(self, texts: List[str],extract_params={}) -> List[List[float]]:        
        embeddings = self._encode(texts,extract_params)
        return embeddings

    def embed_query(self, text: str, extract_params={}) -> List[float]:    
        embedding = self._encode([text],extract_params)
        return embedding[0]

    # copied from https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2#usage-huggingface-transformers
    def get_embedding_with_token_count(self, sentences: Union[str, List[str]], ignore_keys:List[str]=["token_type_ids"]):
        # Mean Pooling - Take attention mask into account for correct averaging
        def mean_pooling(model_output, attention_mask):
            # First element of model_output contains all token embeddings
            token_embeddings = model_output[0]
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
                input_mask_expanded.sum(1), min=1e-9
            )

        # Tokenize sentences
        encoded_input = self.tokenizer(
            sentences, padding=True, truncation=True, return_tensors="pt"
        )
        inputs = encoded_input.to(self.device)
        token_count = inputs["attention_mask"].sum(dim=1).tolist()[0]
        # Compute token embeddings

        for ignore_key in ignore_keys:
            if hasattr(inputs, ignore_key):
                del inputs[ignore_key]

        model_output = self.model(**inputs)
        # Perform pooling
        sentence_embeddings = mean_pooling(model_output, inputs["attention_mask"])
        # Normalize embeddings
        sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)

        return token_count, sentence_embeddings


import whisper
from typing import Dict, List, Tuple
from byzerllm.utils.types import StopSequencesCriteria

from typing import Dict, Any, List, Generator
from pyjava.storage import streaming_tar as STar
import json
import base64
import tempfile
import time


def get_meta(self):
    config = self.config
    return [
        {
            "model_deploy_type": "proprietary",
            "backend": "transformers",
            "max_model_len": getattr(config, "model_max_length", -1),
            "architectures": getattr(config, "architectures", []),
        }
    ]


def stream_chat(
    self,
    tokenizer,
    ins: str,
    his: List[Dict[str, str]] = [],
    max_length: int = 4090,
    top_p: float = 0.95,
    temperature: float = 0.1,
    **kwargs,
):
    audio_input = json.loads(ins)
    audio = audio_input["audio"]
    response_format = audio_input.get("response_format", "json")
    timestamp_granularities = audio_input.get(
        "timestamp_granularities", ["word", "segment"]
    )
    data_prefix = "data:audio/"
    base64_prefix = ";base64,"
    if not audio.startswith(data_prefix):
        raise ValueError("Invalid audio data format")

    format_end = audio.index(base64_prefix)
    audio_format = audio[len(data_prefix) : format_end]
    base64_data = audio[format_end + len(base64_prefix) :]

    # Decode the base64 audio data
    audio_data = base64.b64decode(base64_data)

    # Create a temporary file with the correct extension
    with tempfile.NamedTemporaryFile(
        delete=False, suffix=f".{audio_format}"
    ) as temp_audio_file:
        temp_audio_file.write(audio_data)
        temp_audio_file_path = temp_audio_file.name

    try:
        start_time = time.monotonic()
        word_timestamps = "word" in timestamp_granularities

        with open(temp_audio_file_path, "rb") as audio_file:
            result = self.transcribe(audio_file, word_timestamps=word_timestamps)
        time_cost = time.monotonic() - start_time
        return [
            (
                json.dumps(result, ensure_ascii=True),
                {
                    "metadata": {
                        "request_id": "",
                        "input_tokens_count": 0,
                        "generated_tokens_count": 0,
                        "time_cost": time_cost,
                        "first_token_time": 0,
                        "speed": 0,
                    }
                },
            )
        ]
    finally:
        # Clean up the temporary file
        import os

        os.unlink(temp_audio_file_path)        


def init_model(
    model_dir, infer_params: Dict[str, str] = {}, sys_conf: Dict[str, str] = {}
):
    model = whisper.load_model("large", model_dir)
    import types

    model.stream_chat = types.MethodType(stream_chat, model)
    model.get_meta = types.MethodType(get_meta, model)
    return (model, None)



from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
from ray.util.client.common import ClientActorHandle, ClientObjectRef
import time
import ray
from .. import get_agent_name,run_agent_func,ChatResponse,modify_message_metadata,modify_message_content
from byzerllm.utils.client import default_chat_wrapper,LLMResponse
from byzerllm.apps.agent import Agents
from byzerllm.apps.agent.extensions.preview_file_agent import PreviewFileAgent
from byzerllm.apps.agent.extensions.python_codesandbox_agent import PythonSandboxAgent
from byzerllm.apps.agent.extensions.visualization_agent import VisualizationAgent
from byzerllm.apps.agent.extensions.assistant_agent import AssistantAgent
from byzerllm.apps.agent.extensions.common_agent import CommonAgent
from byzerllm.apps.agent.extensions.spark_sql_agent import SparkSQLAgent
from byzerllm.apps.agent.extensions.rhetorical_agent import RhetoricalAgent
from byzerllm.apps.agent.extensions.sql_reviewer_agent import SQLReviewerAgent
from byzerllm.apps.agent.extensions.byzer_engine_agent import ByzerEngineAgent
from byzerllm.apps.agent.extensions.retrieval_agent import RetrievalAgent
from byzerllm.apps.agent.extensions.llama_index_retrieval_agent import LlamaIndexRetrievalAgent
from byzerllm.apps.agent.extensions.load_data_agent import LoadDataAgent
from byzerllm.apps.agent.store import MessageStore

class DataAnalysisPipeline(ConversableAgent):  
    DEFAULT_SYSTEM_MESSAGE = '''You are a helpful data analysis assistant.
You don't need to write code, or anwser the question. The only thing you need to do 
is plan the data analysis pipeline.

You have following agents to use:

1. visualization_agent, 这个 Agent 可以帮助你对数据进行可视化。
2. assistant_agent, 这个 Agent 可以帮你生成代码对数据进行分析，统计。


Please check the user's question and decide which agent you need to use. And then reply the agent name only.
If there is no agent can help you, 
you should reply exactly `UPDATE CONTEXT`.
''' 

    DEFAULT_USER_MESSAGE = """
"""

    def __init__(
        self,
        name: str,        
        llm: ByzerLLM,
        retrieval: ByzerRetrieval, 
        file_path:str,
        file_ref:ClientObjectRef ,        
        chat_name:str,
        owner:str,
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        message_store:Optional[Optional[Union[str,ClientActorHandle,MessageStore]]]=None,
        **kwargs,
    ):
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,   
            message_store=message_store,         
            **kwargs,
        )
        self.chat_name = chat_name
        self.owner = owner
        self.file_path = file_path
        self.file_ref = file_ref        
        self._reply_func_list = []
        
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], DataAnalysisPipeline.run_pipeline) 
        self.register_reply([Agent, ClientActorHandle,str], DataAnalysisPipeline.reply_agent) 
        self.register_reply([Agent, ClientActorHandle,str], DataAnalysisPipeline.reply_reheorical_agent)         
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 

        params = {}
        if "chat_wrapper" in kwargs:
            params["chat_wrapper"] = kwargs["chat_wrapper"]

        if message_store:
            params["message_store"] = message_store 
            params["group_name"] = self.name   

        max_consecutive_auto_reply = 100000;            

        self.python_interpreter = Agents.create_local_agent(PythonSandboxAgent,"python_interpreter",
                                                llm,retrieval,
                                                chat_name=self.chat_name,
                                                owner=self.owner,
                                                max_consecutive_auto_reply=max_consecutive_auto_reply,
                                                system_message="you are a code sandbox",**params
                                                )
        

        self.preview_file_agent = Agents.create_local_agent(PreviewFileAgent,"privew_file_agent",llm,retrieval,
                                                                chat_name=self.chat_name,
                                                                owner=self.owner,
                                                                max_consecutive_auto_reply=max_consecutive_auto_reply,
                                                                code_agent = self.python_interpreter,**params
                                        )
        
        self.visualization_agent = Agents.create_local_agent(VisualizationAgent,"visualization_agent",llm,retrieval,
                                                            chat_name=self.chat_name,
                                                            owner=self.owner,
                                                            max_consecutive_auto_reply=max_consecutive_auto_reply,                                        
                                                            code_agent = self.python_interpreter,**params
                                        ) 
        self.assistant_agent = Agents.create_local_agent(AssistantAgent,"assistant_agent",llm,retrieval,
                                                        chat_name=self.chat_name,
                                                        owner=self.owner,
                                                        max_consecutive_auto_reply=max_consecutive_auto_reply,
                                                        code_agent = self.python_interpreter,**params)  
        self.common_agent = Agents.create_local_agent(CommonAgent,"common_agent",llm,retrieval,
                                                       chat_name=self.chat_name,
                                                        owner=self.owner,
                                                        max_consecutive_auto_reply=max_consecutive_auto_reply,
                                                        code_agent = self.python_interpreter,**params) 
        
        self.sql_reviewer_agent = Agents.create_local_agent(SQLReviewerAgent,"sql_reviewer_agent",llm,retrieval,chat_name=self.chat_name,
                                                        owner=self.owner,max_consecutive_auto_reply=max_consecutive_auto_reply,**params
                                                            )
        self.byzer_engine_agent = Agents.create_local_agent(ByzerEngineAgent,"byzer_engine_agent",llm,retrieval,chat_name=self.chat_name,
                                                        owner=self.owner,max_consecutive_auto_reply=max_consecutive_auto_reply,**params
                                                            )
        
        self.spark_sql_agent = Agents.create_local_agent(SparkSQLAgent,"spark_sql_agent",llm,retrieval,
                                                         sql_reviewer_agent=self.sql_reviewer_agent,
                                                         byzer_engine_agent=self.byzer_engine_agent,
                                                        chat_name=self.chat_name,
                                                        owner=self.owner,                                                        
                                                        max_consecutive_auto_reply=max_consecutive_auto_reply,**params)   
        self.rhetoorical_agent = Agents.create_local_agent(RhetoricalAgent,"rhetoorical_agent",llm,retrieval,
                                                            chat_name=self.chat_name,
                                                            owner=self.owner,
                                                            max_consecutive_auto_reply=max_consecutive_auto_reply,**params)  
        self.retrieval_agent = Agents.create_local_agent(RetrievalAgent,"retrieval_agent",llm,retrieval,                                   
                                                        chat_name=self.chat_name,
                                                        owner=self.owner,                                                         
                                                        code_agent = None,**params
                                                        )               
        
        self.llama_index_query_agent = Agents.create_local_agent(LlamaIndexRetrievalAgent,"llama_index_query_agent",llm,retrieval,                                   
                                                        chat_name=self.chat_name,
                                                        owner=self.owner,                                                         
                                                        **params
                                                        ) 

        self.load_data_agent = Agents.create_local_agent(LoadDataAgent,"load_data_agent",llm,retrieval,                                   
                                                        chat_name=self.chat_name,
                                                        owner=self.owner,                                                         
                                                        **params
                                                        )   
        
        self.agents = {
            "assistant_agent":self.assistant_agent,
            "visualization_agent":self.visualization_agent,
            "common_agent":self.common_agent,
            "privew_file_agent":self.preview_file_agent,
            "python_interpreter":self.python_interpreter,
            "sql_reviewer_agent":self.sql_reviewer_agent,
            "byzer_engine_agent":self.byzer_engine_agent,            
            "rhetoorical_agent":self.rhetoorical_agent,            
            "spark_sql_agent":self.spark_sql_agent,
            "llama_index_query_agent":self.llama_index_query_agent,
            "load_data_agent":self.load_data_agent,
            "retrieval_agent":self.retrieval_agent
        } 
        self.reply_from_agent = {}       

    def get_agent_chat_messages(self,agent_name:str):
        return self.agents[agent_name].get_chat_messages()
    
    def update_system_message_by_agent(self, agent_name:str,system_message: str):
        if agent_name in self.agents:
            self.agents[agent_name].update_system_message(system_message)
            return True
        return False  
    
    def clear_agent_message_box(self,agent_name:str,last_n:int=0):
        '''
        for debug only
        remove the last n messages from the agent's message box
        '''
        if last_n > 0:
            self._messages[agent_name] = self._messages[agent_name][:-last_n]
            if agent_name in self.agents:
                self.agents[agent_name]._messages[get_agent_name(self)] = self.agents[agent_name]._messages[get_agent_name(self)][:-last_n]
        else:
            self.clear_history(agent_name)
            if agent_name in self.agents:
                self.agents[agent_name].clear_history(get_agent_name(self))                
        return True

    def get_agent_names(self):
        return list(self.agents.keys())  

    def get_agent_system_message(self,agent_name:str):
        return self.agents[agent_name].system_message  

    def send_from_agent_to_agent(self,source_agent:str, target_agent:str, message:Dict):
        '''
        for debug only        
        '''
        self.agents[source_agent].send(message=message,recipient=target_agent)    

    def preview_file(self):
        print(f"send the preview file message to preview_file_agent:{self.file_path}",flush=True)
        self.send(message={
            "content":f"We have a file, the file path is: {self.file_path} , please preview this file",
            "role":"user",
            "metadata":{
                "file_path":self.file_path,
                "file_ref":self.file_ref
            }
        },recipient=self.preview_file_agent)
        
        print("sync the conversation of preview_file_agent to other agents",flush=True)
        for agent in [self.visualization_agent,self.assistant_agent,self.python_interpreter]:            
            for message in self._messages["privew_file_agent"]:                 
                self.send(message=message,recipient=agent,request_reply=False)

    def select_agent(self,raw_message,messages,sender):
        _,llm_reply = self.generate_llm_reply(raw_message,messages,sender)
        fail = "UPDATE CONTEXT" in llm_reply[-20:].upper() or "UPDATE CONTEXT" in llm_reply[:20].upper()
        if fail:
            return True, None
        else:
            return True,llm_reply.strip()

    def get_agent_stream_messages(self,agent_name:str,id:str):
        for message in self.agents[agent_name]._stream_get_message_from_self(id):
            yield message
        
    def reply_reheorical_agent(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:
        if get_agent_name(sender) != "rhetoorical_agent":
            return False,None                            
        return True,None
    
    def reply_agent(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:
        if get_agent_name(sender) not in self.agents:
            return False,None                            
        return True,None
                    

    def run_pipeline(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:
                
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
        
        ori_message = messages[-1]

                                         
        # self.send(message=ori_message,recipient=self.rhetoorical_agent)

        specified_agent = ori_message.get("metadata",{}).get("agent",None)  

        agent_name = None
        if specified_agent:
            agent_name = specified_agent
        else:        
            _,_agent_name = self.select_agent(raw_message,messages,sender)
            agent_name = _agent_name
        
        print(f"Select agent: {agent_name} to answer the question: {ori_message['content'][0:20]}",flush=True)
        
        if agent_name and agent_name in self.agents:
            agent = self.agents[agent_name]
            # reset the agent except the conversation history  
            self._prepare_chat(agent, clear_history=False)                      
            self.send(message=ori_message,recipient=agent)                                                
            agent_reply = self._messages[get_agent_name(agent)][-1]
            if isinstance(agent_reply,dict):
                return True,agent_reply
            return True, {"content":agent_reply}
        
        return self.generate_llm_reply(raw_message,messages,sender)

class DataAnalysisPipelineManager:
    def __init__(self) -> None:
        self.pipelines = {}
        self.lasted_updated = {}
    
    ## if the sandbox is not used for 1h, we will remove it
    def check_pipeline_timeout(self,timeout:int=60*60): 
        remove_names = []
        for name in self.lasted_updated:
            if time.time() - self.lasted_updated[name] > timeout:
                remove_names.append(name)
        for name in remove_names:
            try:
                del self.pipelines[name]                
            except:
                pass
            try:
                del self.lasted_updated[name]
            except:
                pass                    

    def check_pipeline_exists(self,name:str)->bool:
        self.check_pipeline_timeout() 
        return name in self.pipelines

    def get_pipeline(self,name:str):                
        self.check_pipeline_timeout()        
        return self.pipelines[name]
    
    def force_clear(self):
        self.pipelines = {}
        self.lasted_updated = {}

    def remove_pipeline(self,name:str):
        if name in self.pipelines:
            try:
                del self.pipelines[name]                
            except:
                pass
            try:
                del self.lasted_updated[name]
            except:
                pass    

    def get_or_create_pipeline( self,
                                name:str,
                                llm:ByzerLLM,retrieval:ByzerRetrieval,
                                file_path:str,
                                file_ref:ClientObjectRef,
                                chat_name:str,
                                owner:str,
                                chat_wrapper:Optional[Callable[[ByzerLLM,Optional[List[Dict]],Dict],List[LLMResponse]]] = None,
                                message_store:Optional[Optional[Union[str,ClientActorHandle,MessageStore]]]=None,
                                num_gpus:int=0,num_cpus:int=0):
        self.lasted_updated[name] = time.time()
        
        self.check_pipeline_timeout()
        if name in self.pipelines:            
            return self.pipelines[name]
        
        pipeline = ray.remote(DataAnalysisPipeline).options(
                name=name,                                              
                num_cpus=num_cpus,
                num_gpus=num_gpus
            ).remote(
                name = name,
                llm = llm,
                retrieval = retrieval,
                file_path=file_path,
                file_ref=file_ref,
                chat_wrapper=chat_wrapper,
                chat_name=chat_name,
                owner=owner ,
                message_store= message_store
                )
        self.pipelines[name] = pipeline
        return pipeline
    


from ..conversable_agent import ConversableAgent
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from ....utils.client import ByzerLLM
from byzerllm.utils.retrieval import ByzerRetrieval
from ..agent import Agent
import ray
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from .. import get_agent_name,run_agent_func,ChatResponse
from byzerllm.apps.agent.extensions.simple_retrieval_client import SimpleRetrievalClient
import uuid
import json
from langchain import PromptTemplate

try:
    from termcolor import colored
except ImportError:

    def colored(x, *args, **kwargs):
        return x

import jieba     



class RetrievalAgent(ConversableAgent): 
    PROMPT_DEFAULT = """You're a retrieve augmented chatbot. You answer user's questions based on your own knowledge and the
context provided by the user. You should follow the following steps to answer a question:
Step 1, you estimate the user's intent based on the question and context. The intent can be a code generation task or
a question answering task.
Step 2, you reply based on the intent.
If you can't answer the question with or without the current context, you should reply exactly `UPDATE CONTEXT`.
If user's intent is code generation, you must obey the following rules:
Rule 1. You MUST NOT install any packages because all the packages needed are already installed.
Rule 2. You must follow the formats below to write your code:
```language
# your code
```

If user's intent is question answering, you must give as short an answer as possible.

User's question is: {input_question}

Context is: {input_context}
"""

    PROMPT_CODE = """You're a retrieve augmented coding assistant. You answer user's questions based on your own knowledge and the
    context provided by the user.
    If you can't answer the question with or without the current context, you should reply exactly `UPDATE CONTEXT`.
    For code generation, you must obey the following rules:
    Rule 1. You MUST NOT install any packages because all the packages needed are already installed.
    Rule 2. You must follow the formats below to write your code:
    ```language
    # your code
    ```

    User's question is: {input_question}

    Context is: {input_context}
    """

    PROMPT_QA = """You're a retrieve augmented chatbot. You answer user's questions based on your own knowledge and the
    context provided by the user.
    If you can't answer the question with or without the current context, you should reply exactly `UPDATE CONTEXT`.
    You must give as short an answer as possible.
    """   

    DEFAULT_SYSTEM_MESSAGE = PROMPT_QA
    
    def __init__(
        self,
        name: str,
        llm: ByzerLLM,        
        retrieval: ByzerRetrieval,        
        chat_name:str,
        owner:str,        
        code_agent: Union[Agent, ClientActorHandle,str],
        byzer_engine_url: str="http://127.0.0.1:9003/model/predict",        
        retrieval_cluster:str="data_analysis",
        retrieval_db:str="data_analysis",
        update_context_retry: int = 3,
        chunk_size_in_context: int = 1,
        system_message: Optional[str] = DEFAULT_SYSTEM_MESSAGE,        
        is_termination_msg: Optional[Callable[[Dict], bool]] = None,
        max_consecutive_auto_reply: Optional[int] = None,
        human_input_mode: Optional[str] = "NEVER",
        code_execution_config: Optional[Union[Dict, bool]] = False,
        **kwargs,
    ):       
        super().__init__(
            name,
            llm,retrieval,
            system_message,
            is_termination_msg,
            max_consecutive_auto_reply,
            human_input_mode,
            code_execution_config=code_execution_config,            
            **kwargs,
        )
        self.chat_name = chat_name
        self.owner = owner
        self.code_agent = code_agent

        self.byzer_engine_url = byzer_engine_url

        self.update_context_retry = update_context_retry
        self.chunk_size_in_context = chunk_size_in_context

        self._reply_func_list = []
        # self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.generate_llm_reply)   
        self.register_reply([Agent, ClientActorHandle,str], RetrievalAgent.generate_retrieval_based_reply) 
        self.register_reply([Agent, ClientActorHandle,str], ConversableAgent.check_termination_and_human_reply) 
                
        self.retrieval_cluster = retrieval_cluster
        self.retrieval_db = retrieval_db 

        self.simple_retrieval_client = SimpleRetrievalClient(llm=self.llm,
                                                             retrieval=self.retrieval,
                                                             retrieval_cluster=self.retrieval_cluster,
                                                             retrieval_db=self.retrieval_db,
                                                             )         
        
                        

    def generate_retrieval_based_reply(
        self,
        raw_message: Optional[Union[Dict,str,ChatResponse]] = None,
        messages: Optional[List[Dict]] = None,
        sender: Optional[Union[ClientActorHandle,Agent,str]] = None,
        config: Optional[Any] = None,
    ) -> Tuple[bool, Union[str, Dict, None,ChatResponse]]:  
        
        if messages is None:
            messages = self._messages[get_agent_name(sender)]
                
        new_message = messages[-1]

        if "file_path" in new_message["metadata"]:
            file_path = new_message["metadata"]["file_path"]
            content = open(file_path,"r").read()
            self.simple_retrieval_client.save_text_content(owner=self.owner,title="",content=content,url=file_path) 

        if "file_ref" in new_message["metadata"]:
            file_ref = new_message["metadata"]["file_ref"]
            file_path = new_message["metadata"].get("file_path","")
            content = ray.get(file_ref)
            self.simple_retrieval_client.save_text_content(owner=self.owner,title="",content=content,url=file_path) 
        
        top_k = 4
        if "top_k" in new_message["metadata"]:
            top_k = new_message["metadata"]["top_k"]    
        
        contents = self.simple_retrieval_client.search_content_chunks(owner=self.owner,q=new_message["content"],limit=top_k,return_json=False)
        for item in contents:
            temp = self.simple_retrieval_client.get_doc(item["doc_id"],owner=self.owner)
            item["doc_url"] = temp["url"]                        

        input_context = json.dumps([{"content":x["raw_chunk"]} for x in contents],ensure_ascii=False,indent=4)

        prompt = PromptTemplate.from_template('''User's question is: {input_question}

Context is: 

```json                                                                                            
{input_context}
```
''').format(input_question=new_message["content"],input_context=input_context)
        
        new_message = {"content":prompt,"role":"user"}
        id = str(uuid.uuid4())        
        v = self.llm.stream_chat_oai(conversations=self._system_message + [new_message])
        
        return self.stream_reply(v,contexts=contents)          
        
                
        
                    


import time
from typing import List, Tuple, Dict, Any, Union
import httpx
from openai import OpenAI
import base64
import io
import json
import ray
from byzerllm.utils.types import (
    BlockVLLMStreamServer,
    StreamOutputs,
    SingleOutput,
    SingleOutputMeta,
    BlockBinaryStreamServer,
)
from byzerllm.utils.langutil import asyncfy_with_semaphore
import threading
import asyncio
import traceback
import uuid
import tempfile        

class CustomSaasAPI:

    def __init__(self, infer_params: Dict[str, str]) -> None:

        self.api_key = infer_params["saas.api_key"]
        self.model = infer_params.get("saas.model", "gpt-3.5-turbo-1106")

        other_params = {}

        if "saas.api_base" in infer_params:
            other_params["api_base"] = infer_params["saas.api_base"]

        if "saas.api_version" in infer_params:
            other_params["api_version"] = infer_params["saas.api_version"]

        if "saas.api_type" in infer_params:
            other_params["api_type"] = infer_params["saas.api_type"]

        if "saas.base_url" in infer_params:
            other_params["base_url"] = infer_params["saas.base_url"]

        if "saas.timeout" in infer_params:
            other_params["timeout"] = float(infer_params["saas.timeout"])

        self.max_retries = int(infer_params.get("saas.max_retries", 10))

        self.meta = {
            "model_deploy_type": "saas",
            "backend": "saas",
            "support_stream": True,
            "model_name": self.model,
        }

        self.meta["embedding_mode"] = "embedding" in self.model.lower()

        self.proxies = infer_params.get("saas.proxies", None)
        self.local_address = infer_params.get("saas.local_address", "0.0.0.0")

        if not self.proxies:
            self.client = OpenAI(**other_params, api_key=self.api_key)
        else:
            self.client = OpenAI(
                **other_params,
                api_key=self.api_key,
                http_client=httpx.Client(
                    proxies=self.proxies,
                    transport=httpx.HTTPTransport(local_address=self.local_address),
                ),
            )

        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError:
            try:
                ray.remote(BlockVLLMStreamServer).options(
                    name="BLOCK_VLLM_STREAM_SERVER",
                    lifetime="detached",
                    max_concurrency=1000,
                ).remote()
            except Exception as e:
                pass
        try:
            ray.get_actor("BlockBinaryStreamServer")
        except ValueError:
            try:
                ray.remote(BlockBinaryStreamServer).options(
                    name="BlockBinaryStreamServer",
                    lifetime="detached",
                    max_concurrency=1000,
                ).remote()
            except Exception as e:
                pass

    # saas/proprietary
    def get_meta(self):
        return [self.meta]

    def process_input(self, ins: Union[str, List[Dict[str, Any]], Dict[str, Any]]):

        if isinstance(ins, list) or isinstance(ins, dict):
            return ins

        content = []
        try:
            ins_json = json.loads(ins)
        except:
            return ins

        ## speech
        if isinstance(ins_json, dict):
            return ins_json

        content = []
        #     [
        #     {"type": "text", "text": "What’s in this image?"},
        #     {
        #       "type": "image_url",
        #       "image_url": {
        #         "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
        #         "detail": "high"
        #       },
        #     },
        #   ],
        for item in ins_json:
            # for format like this: {"image": "xxxxxx", "text": "What’s in this image?","detail":"high"}
            # or {"image": "xxxxxx"}, {"text": "What’s in this image?"}
            if ("image" in item or "image_url" in item) and "type" not in item:
                image_data = item.get("image", item.get("image_url", ""))
                ## "data:image/jpeg;base64,"
                if not image_data.startswith("data:"):
                    image_data = "data:image/jpeg;base64," + image_data

                ## get the other fields except image/text/image_url
                other_fields = {
                    k: v
                    for k, v in item.items()
                    if k not in ["image", "text", "image_url"]
                }
                content.append(
                    {
                        "image_url": {"url": image_data, **other_fields},
                        "type": "image_url",
                    }
                )            

            if "text" in item and "type" not in item:
                text_data = item["text"]
                content.append({"text": text_data, "type": "text"})

            ## for format like this: {"type": "text", "text": "What’s in this image?"},
            ## {"type": "image_url", "image_url": {"url":"","detail":"high"}}
            ## this is the standard format, just return it
            if "type" in item and item["type"] == "text":
                content.append(item)

            if "type" in item and item["type"] == "image_url":
                content.append(item)

        if not content:
            return ins

        return content

    async def async_embed_query(self, ins: str, **kwargs):
        resp = await asyncfy_with_semaphore(
            lambda: self.client.embeddings.create(input=[ins], model=self.model)
        )()
        embedding = resp.data[0].embedding
        usage = resp.usage
        return (
            embedding,
            {
                "metadata": {
                    "input_tokens_count": usage.prompt_tokens,
                    "generated_tokens_count": 0,
                }
            },
        )

    def embed_query(self, ins: str, **kwargs):
        resp = self.client.embeddings.create(input=[ins], model=self.model)
        embedding = resp.data[0].embedding
        usage = resp.usage
        return (
            embedding,
            {
                "metadata": {
                    "input_tokens_count": usage.prompt_tokens,
                    "generated_tokens_count": 0,
                }
            },
        )

    async def async_text_to_speech(
        self, stream: bool, ins: str, voice: str, chunk_size: int = None, **kwargs
    ):
        if stream:
            server = ray.get_actor("BlockBinaryStreamServer")
            request_id = [None]

            def writer():
                try:
                    request_id[0] = str(uuid.uuid4())
                    with self.client.with_streaming_response.audio.speech.create(
                        model=self.model, voice=voice, input=ins, **kwargs
                    ) as response:
                        for chunk in response.iter_bytes(chunk_size):
                            input_tokens_count = 0
                            generated_tokens_count = 0
                            ray.get(
                                server.add_item.remote(
                                    request_id[0],
                                    StreamOutputs(
                                        outputs=[
                                            SingleOutput(
                                                text=chunk,
                                                metadata=SingleOutputMeta(
                                                    input_tokens_count=input_tokens_count,
                                                    generated_tokens_count=generated_tokens_count,
                                                ),
                                            )
                                        ]
                                    ),
                                )
                            )
                except:
                    traceback.print_exc()
                ray.get(server.mark_done.remote(request_id[0]))

            threading.Thread(target=writer, daemon=True).start()

            time_count = 10 * 100
            while request_id[0] is None and time_count > 0:
                time.sleep(0.01)
                time_count -= 1

            if request_id[0] is None:
                raise Exception("Failed to get request id")

            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))

            await asyncio.to_thread(write_running)
            return [
                (
                    "",
                    {
                        "metadata": {
                            "request_id": request_id[0],
                            "stream_server": "BlockBinaryStreamServer",
                        }
                    },
                )
            ]

        start_time = time.monotonic()
        with io.BytesIO() as output:
            async with asyncfy_with_semaphore(
                lambda: self.client.with_streaming_response.audio.speech.create(
                    model=self.model, voice=voice, input=ins, **kwargs
                )
            )() as response:
                for chunk in response.iter_bytes():
                    output.write(chunk)

            base64_audio = base64.b64encode(output.getvalue()).decode()
            time_cost = time.monotonic() - start_time
            return [
                (
                    base64_audio,
                    {
                        "metadata": {
                            "request_id": "",
                            "input_tokens_count": 0,
                            "generated_tokens_count": 0,
                            "time_cost": time_cost,
                            "first_token_time": 0,
                            "speed": 0,
                        }
                    },
                )
            ]

    async def async_speech_to_text(
        self,
        audio: str,
        response_format: str = "verbose_json",
        timestamp_granularities: List[str] = ["word","segment"],
    ):               
        # Extract audio format and base64 data
        data_prefix = "data:audio/"
        base64_prefix = ";base64,"
        if not audio.startswith(data_prefix):
            raise ValueError("Invalid audio data format")
        
        format_end = audio.index(base64_prefix)
        audio_format = audio[len(data_prefix):format_end]
        base64_data = audio[format_end + len(base64_prefix):]
        
        # Decode the base64 audio data
        audio_data = base64.b64decode(base64_data)

        # Create a temporary file with the correct extension
        with tempfile.NamedTemporaryFile(delete=False, suffix=f'.{audio_format}') as temp_audio_file:
            temp_audio_file.write(audio_data)
            temp_audio_file_path = temp_audio_file.name

        try:
            start_time = time.monotonic()
                   
            with open(temp_audio_file_path, "rb") as audio_file:                
                transcription = self.client.audio.transcriptions.create(
                    model=self.model or "whisper-1",
                    file=audio_file,
                    response_format=response_format,
                    timestamp_granularities=timestamp_granularities,
                )
            time_cost = time.monotonic() - start_time

            return [
                (
                    json.dumps(transcription.to_dict(), ensure_ascii=True),
                    {
                        "metadata": {
                            "request_id": "",
                            "input_tokens_count": 0,
                            "generated_tokens_count": 0,
                            "time_cost": time_cost,
                            "first_token_time": 0,
                            "speed": 0,
                        }
                    },
                )
            ]
        finally:
            # Clean up the temporary file
            import os
            os.unlink(temp_audio_file_path)

    def image_to_text(self, ins: str, **kwargs):
        pass

    async def async_text_to_image(
        self, stream: bool, input: str, size: str, quality: str, n: int, **kwargs
    ):
        if stream:
            raise Exception("Stream not supported for text to image")
        start_time = time.monotonic()
        response = await asyncfy_with_semaphore(
            lambda: self.client.images.generate(
                model=self.model,
                prompt=input,
                size=size,
                quality=quality,
                n=1,
                response_format="b64_json",
                **kwargs,
            )
        )()
        time_cost = time.monotonic() - start_time
        base64_image = response.data[0].b64_json
        return [
            (
                base64_image,
                {
                    "metadata": {
                        "request_id": "",
                        "input_tokens_count": 0,
                        "generated_tokens_count": 0,
                        "time_cost": time_cost,
                        "first_token_time": 0,
                        "speed": 0,
                    }
                },
            )
        ]

    def text_to_text(self, ins: str, **kwargs):
        pass

    async def async_stream_chat(
        self,
        tokenizer,
        ins: str,
        his: List[Dict[str, Any]] = [],
        max_length: int = 4096,
        top_p: float = 0.7,
        temperature: float = 0.9,
        **kwargs,
    ):

        model = self.model

        if "model" in kwargs:
            model = kwargs["model"]

        messages = [
            {"role": message["role"], "content": self.process_input(message["content"])}
            for message in his
        ] + [{"role": "user", "content": self.process_input(ins)}]

        stream = kwargs.get("stream", False)

        ## content = [
        ##    "voice": "alloy","input": "Hello, World!",response_format: "mp3"]
        last_message = messages[-1]["content"]

        if isinstance(last_message, dict) and "voice" in last_message:
            voice = last_message["voice"]
            response_format = last_message.get("response_format", "mp3")
            chunk_size = last_message.get("chunk_size", None)
            input = last_message["input"]
            return await self.async_text_to_speech(
                stream=stream,
                ins=input,
                voice=voice,
                chunk_size=chunk_size,
                response_format=response_format,
            )

        if isinstance(last_message, dict) and "input" in last_message:
            input = last_message["input"]
            size = last_message.get("size", "1024x1024")
            quality = last_message.get("quality", "standard")
            n = last_message.get("n", 1)
            return await self.async_text_to_image(
                stream=stream, input=input, size=size, quality=quality, n=n
            )

        if isinstance(last_message, dict) and "audio" in last_message:            
            audio_data = last_message.get("audio", "")            
            tpe = last_message.get("type","wav")
            if not audio_data.startswith("data:"):
                audio_data = f"data:audio/${tpe};base64," + audio_data                
            return await self.async_speech_to_text(audio=audio_data)
        
        server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        request_id = [None]

        def writer():
            try:
                r = ""
                response = self.client.chat.completions.create(
                    messages=messages,
                    model=model,
                    stream=True,
                    max_tokens=max_length,
                    temperature=temperature,
                    top_p=top_p,
                )
                # input_tokens_count = 0
                # generated_tokens_count = 0

                request_id[0] = str(uuid.uuid4())

                for chunk in response:
                    content = chunk.choices[0].delta.content or ""
                    r += content
                    if hasattr(chunk, "usage") and chunk.usage:
                        input_tokens_count = chunk.usage.prompt_tokens
                        generated_tokens_count = chunk.usage.completion_tokens
                    else:
                        input_tokens_count = 0
                        generated_tokens_count = 0
                    ray.get(
                        server.add_item.remote(
                            request_id[0],
                            StreamOutputs(
                                outputs=[
                                    SingleOutput(
                                        text=r,
                                        metadata=SingleOutputMeta(
                                            input_tokens_count=input_tokens_count,
                                            generated_tokens_count=generated_tokens_count,
                                        ),
                                    )
                                ]
                            ),
                        )
                    )
            except:
                traceback.print_exc()
            ray.get(server.mark_done.remote(request_id[0]))

        if stream:
            threading.Thread(target=writer, daemon=True).start()

            time_count = 10 * 100
            while request_id[0] is None and time_count > 0:
                time.sleep(0.01)
                time_count -= 1

            if request_id[0] is None:
                raise Exception("Failed to get request id")

            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))

            await asyncio.to_thread(write_running)
            return [
                (
                    "",
                    {
                        "metadata": {
                            "request_id": request_id[0],
                            "stream_server": "BLOCK_VLLM_STREAM_SERVER",
                        }
                    },
                )
            ]
        else:
            try:
                start_time = time.monotonic()
                response = await asyncfy_with_semaphore(
                    lambda: self.client.chat.completions.create(
                        messages=messages,
                        model=model,
                        max_tokens=max_length,
                        temperature=temperature,
                        top_p=top_p,
                    )
                )()

                generated_text = response.choices[0].message.content
                generated_tokens_count = response.usage.completion_tokens
                input_tokens_count = response.usage.prompt_tokens
                time_cost = time.monotonic() - start_time
                return [
                    (
                        generated_text,
                        {
                            "metadata": {
                                "request_id": response.id,
                                "input_tokens_count": input_tokens_count,
                                "generated_tokens_count": generated_tokens_count,
                                "time_cost": time_cost,
                                "first_token_time": 0,
                                "speed": float(generated_tokens_count) / time_cost,
                            }
                        },
                    )
                ]
            except Exception as e:
                print(f"Error: {e}")
                raise e



from typing import List, Dict
import time
import qianfan
import threading
import asyncio
import ray

from byzerllm.utils import random_uuid
from byzerllm.log import init_logger
from byzerllm.utils.types import BlockVLLMStreamServer, StreamOutputs, SingleOutput, SingleOutputMeta
from byzerllm.utils.langutil import asyncfy_with_semaphore

logger = init_logger(__name__)


class CustomSaasAPI:
    def __init__(self, infer_params: Dict[str, str]) -> None:
        self.api_key: str = infer_params.get("saas.api_key", "")
        self.secret_key: str = infer_params.get("saas.secret_key", "")
        self.access_token: str = infer_params.get("saas.access_token", "")

        if not self.access_token and (not self.api_key or not self.secret_key):
            raise ValueError("Please specify either access_token or ak/sk")

        # os.environ["QIANFAN_ACCESS_KEY"] = self.api_key
        # os.environ["QIANFAN_SECRET_KEY"] = self.secret_key
        # qianfan.AK(self.api_key)
        # qianfan.SK(self.secret_key)
        self.model: str = infer_params.get("saas.model", "ERNIE-Bot-turbo")
        self.client = qianfan.ChatCompletion(ak=self.api_key, sk=self.secret_key, access_token=self.access_token)
        try:
            ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
        except ValueError:            
            ray.remote(BlockVLLMStreamServer).options(name="BLOCK_VLLM_STREAM_SERVER",lifetime="detached",max_concurrency=1000).remote() 

     # saas/proprietary
    def get_meta(self):
        return [{
            "model_deploy_type": "saas",
            "backend":"saas",
            "support_stream": True
        }] 

    async def async_get_meta(self):
        return await asyncfy_with_semaphore(self.get_meta)()   

    async def async_stream_chat(
            self,
            tokenizer,
            ins: str,
            his: List[dict] = [],
            max_length: int = 4096,
            top_p: float = 0.7,
            temperature: float = 0.9,
            **kwargs
    ):
        request_id = kwargs.get("request_id", random_uuid())
        other_params = {}
        
        if "request_timeout" in kwargs:
            other_params["request_timeout"] = int(kwargs["request_timeout"])
        
        if "retry_count" in kwargs:
            other_params["retry_count"] = int(kwargs["retry_count"])
        
        if "backoff_factor" in kwargs:
            other_params["backoff_factor"] = float(kwargs["backoff_factor"])  

        if "penalty_score" in kwargs:
            other_params["penalty_score"] = float(kwargs["penalty_score"])  

        stream = kwargs.get("stream",False)                                  

        messages = qianfan.Messages()
        for item in his:
            role, content = item['role'], item['content']
            # messages must have an odd number of members
            # look for details: https://cloud.baidu.com/doc/WENXINWORKSHOP/s/clntwmv7t
            if role == 'system':
                messages.append(content, qianfan.Role.User)
                messages.append("OK", qianfan.Role.Assistant)
                continue
            messages.append(content, role)

        if ins:
            messages.append(ins, qianfan.Role.User)
        
        start_time = time.monotonic()

        logger.info(f"Receiving request {request_id} model: {self.model}")

        res_data = await asyncfy_with_semaphore(lambda:self.client.do(
            model=self.model,            
            messages=messages,
            top_p=top_p,
            temperature=temperature,
            stream=stream,
            **other_params
        ))()
        
        if stream:
            server = ray.get_actor("BLOCK_VLLM_STREAM_SERVER")
            request_id = [None]

            def writer(): 
                for response in res_data:                                        
                    if response["code"] == 200:
                        v = response["result"]
                        request_id[0] = f'qianfan_{response["id"]}'
                        ray.get(server.add_item.remote(request_id[0], 
                                                       StreamOutputs(outputs=[SingleOutput(text=v,metadata=SingleOutputMeta(
                                                           input_tokens_count=response["usage"]["prompt_tokens"],
                                                           generated_tokens_count=response["usage"]["completion_tokens"],
                                                       ))])
                                                       ))                                            
                ray.get(server.mark_done.remote(request_id[0]))

            threading.Thread(target=writer,daemon=True).start()            
                               
            time_count= 10*100
            while request_id[0] is None and time_count > 0:
                time.sleep(0.01)
                time_count -= 1
            
            if request_id[0] is None:
                raise Exception("Failed to get request id")
            
            def write_running():
                return ray.get(server.add_item.remote(request_id[0], "RUNNING"))
                        
            await asyncio.to_thread(write_running)
            return [("",{"metadata":{"request_id":request_id[0],"stream_server":"BLOCK_VLLM_STREAM_SERVER"}})] 

        time_cost = time.monotonic() - start_time

        generated_text = res_data["result"]
        generated_tokens_count = res_data["usage"]["completion_tokens"]
        input_tokens_count = res_data["usage"]["prompt_tokens"]

        logger.info(
            f"Completed request {request_id} "
            f"model: {self.model} "
            f"cost: {time_cost} "
            f"result: {res_data}"
        )

        return [(generated_text,{"metadata":{
                "request_id":res_data["id"],
                "input_tokens_count":input_tokens_count,
                "generated_tokens_count":generated_tokens_count,
                "time_cost":time_cost,
                "first_token_time":0,
                "speed":float(generated_tokens_count)/time_cost,        
            }})] 


import ray
from pyjava import RayContext
from pyjava.udf import UDFMaster,UDFWorker,UDFBuilder,UDFBuildInFunc
from typing import Any, NoReturn, Callable, Dict, List
from ray.util.client.common import ClientActorHandle, ClientObjectRef
from byzerllm import consume_model
from byzerllm.records import SearchQuery
from byzerllm.utils.retrieval import ByzerRetrieval
from byzerllm.records import SearchQuery
import json

def search_func(model,v):
    data = [json.loads(item) for item in v]
    results=[]
    for item in data:
        cluster_name = item["clusterName"]
        database = item["database"]
        table = item["table"]
        vector = []
        if "query.vector" in item:
            vector_str = item["query.vector"] 
            if vector_str.startswith("["):
                vector = json.loads(vector_str)
            else:
                vector = [float(i) for i in vector_str.split(",")]

        fields = []
        if "query.fields" in item:
            fields_str = item["query.fields"]
            if fields_str.startswith("["):
                fields = json.loads(fields_str)
            else:
                fields = fields_str.split(",")

        keyword = None
        if "query.keyword"  in item:
            keyword = item["query.keyword"]

        vector_field = None
        if "query.vectorField" in item:
            vector_field = item["query.vectorField"]    

        query = SearchQuery(
            keyword=keyword,
            fields=fields,
            vector=vector,
            vectorField=vector_field,
            limit=int(item.get("query.limit",10)),
        )
        docs = model.search(cluster_name,database,table,query)
        results.append(docs)
    return {"value":[json.dumps(docs,ensure_ascii=False)]}
                 
def init_retrieval_client(model_refs: List[ClientObjectRef], conf: Dict[str, str]) -> Any:
    consume_model(conf)
    byzer = ByzerRetrieval()
    byzer.launch_gateway()
    return byzer




from sentence_transformers import SentenceTransformer
from typing import Dict,List,Tuple

def _encode(self,texts: List[str],extract_params={}):        
    embeddings = [emb.tolist() for emb in self.encode(texts,
                                                      normalize_embeddings=extract_params.get("normalize_embeddings",True))]
    return embeddings
    
def embed_documents(self, texts: List[str],extract_params={}) -> List[List[float]]:        
    embeddings = self._encode(texts,extract_params)
    return embeddings

def embed_query(self, text: str,extract_params={}) -> List[float]:    
    embedding = self._encode([text],extract_params)
    return embedding[0]
    

def init_model(model_dir,infer_params,sys_conf={}):        
    model = SentenceTransformer(model_dir) 
    import types
    model._encode = types.MethodType(_encode, model) 
    model.embed_documents = types.MethodType(embed_documents, model) 
    model.embed_query = types.MethodType(embed_query, model)     
    return (None,model)





import asyncio
from typing import List, Dict, Any

import ray.experimental
from byzerllm.apps.byzer_storage.simple_api import ByzerStorage
import time
import json
import os
import concurrent.futures
import io
import sys
from contextlib import redirect_stdout, redirect_stderr
import queue
import threading
import ray
import byzerllm
from typing import List, Union, Dict
from pydantic import BaseModel
from byzerllm.apps.utils import TagExtractor, Tag
import importlib
from loguru import logger


class QAPair(BaseModel):
    question: str
    answer: str


def read_alpaca_zh():
    with importlib.resources.path(
        "byzerllm.apps.byzer_storage", "alpaca_zh.json"
    ) as json_path:
        with open(json_path, "r", encoding="utf-8") as f:
            return json.load(f)


class MemoryManager:
    _queue = asyncio.Queue()
    _is_processing = False

    def __init__(
        self,
        storage: ByzerStorage,
        base_dir: str,
        model_name: str = "deepseek_chat",
        remote: bool = True,
    ):
        self.storage = storage
        home = os.path.expanduser("~")
        self.base_dir = base_dir or os.path.join(home, ".auto-coder")
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=5)
        self.log_file = None
        self.remote = remote
        self.model_name = model_name

    @classmethod
    async def add_to_queue(cls, name: str, memories: List[str]):
        await cls._queue.put((name, memories))
        if not cls._is_processing:
            asyncio.create_task(cls.process_queue())

    @classmethod
    async def process_queue(cls):
        cls._is_processing = True
        while not cls._queue.empty():
            name, memories = await cls._queue.get()
            instance = cls.get_instance()
            await instance.memorize(name, memories)
            cls._queue.task_done()
        cls._is_processing = False

    async def memorize(
        self, name: str, memories: List[str], options: Dict[str, Any] = {}
    ):
        try:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                self.thread_pool, self._memorize_with_logs, name, memories, options
            )
        finally:
            if self.remote:
                ray.actor.exit_actor()

        print(f"Memorization for {name} completed.")

    def _memorize_with_logs(
        self, name: str, memories: List[str], options: Dict[str, Any] = {}
    ):

        if self.remote:
            self._memorize(name, memories, options=options)
            return

        logs_dir = os.path.join(self.base_dir, "storage", "logs", "memorize")
        os.makedirs(logs_dir, exist_ok=True)
        log_file = os.path.join(logs_dir, f"{name}.log")

        log_queue = queue.Queue()
        stop_event = threading.Event()

        # Start the log writer thread
        log_writer_thread = threading.Thread(
            target=self._log_writer, args=(log_queue, log_file, stop_event)
        )
        self.log_file = log_file
        log_writer_thread.start()

        class QueueStream:
            def __init__(self, queue):
                self.queue = queue

            def write(self, msg):
                self.queue.put(msg)

            def flush(self):
                pass

        queue_stream = QueueStream(log_queue)

        with redirect_stdout(queue_stream), redirect_stderr(queue_stream):
            self._memorize(name, memories)

        # Signal the log writer to stop and wait for it to finish
        stop_event.set()
        log_writer_thread.join()

    def _log_writer(
        self, log_queue: queue.Queue, log_file: str, stop_event: threading.Event
    ):
        with open(log_file, "w") as f:
            while not stop_event.is_set() or not log_queue.empty():
                try:
                    msg = log_queue.get(timeout=0.1)
                    f.write(msg)
                    f.flush()
                except queue.Empty:
                    continue

    @byzerllm.prompt()
    def _convert_pretrain_text_to_instruction(self, text: str) -> str:
        """
        根据提供的信息，生成多个相关的问题，这些问题的回答，最终要能覆盖里面所有的信息,生成的问题越多越好。

        下面是一些信息：

        {{text}}

        现在请开始生成问题，每个问题使用<_question_>标签包裹，每个回答使用<_answer_>标签包裹，最后
        每个一组问题和回答使用<_group_>标签包裹。
        """

    @byzerllm.prompt()
    def _format(self, text: str) -> str:
        """
        下面是一些问答信息：

        {{text}}


        请将每个问题使用<_question_>标签包裹，每个回答使用<_answer_>标签包裹，最后
        每个一组问题和回答使用<_group_>标签包裹。
        """

    def to_qa_pairs(self, text: str, llm) -> List[QAPair]:
        print(f"Generating QA pairs for {text}", flush=True)
        v = self._convert_pretrain_text_to_instruction.with_llm(llm).run(text)
        # format_v = self._format.with_llm(llm).run(v)
        root_tag = TagExtractor(v).extract()
        qa_pairs = []
        # _group_
        for item in root_tag.content:
            qas = item.content
            if len(qas) == 2:
                if (
                    qas[0].start_tag == "<_question_>"
                    and qas[0].end_tag == "</_question_>"
                    and qas[1].start_tag == "<_answer_>"
                    and qas[1].end_tag == "</_answer_>"
                ):
                    qa_pairs.append(
                        QAPair(question=qas[0].content, answer=qas[1].content)
                    )

        print(f"Generated {len(qa_pairs)} QA pairs.", flush=True)
        logger.info(f"Generated {len(qa_pairs)} QA pairs.")
        return qa_pairs

    def _memorize(
        self,
        name: str,
        memories: List[Union[str, Dict[str, Any]]],
        options: Dict[str, Any] = {},
    ):
        data = []
        stage = options.get("stage", "pt")
        min_samples = options.pop("min_samples", 1000)
        base_model_dir = os.path.join(self.base_dir, "storage", "models")
        llama_model = os.path.join(
            base_model_dir, "meta-llama", "Meta-Llama-3-8B-Instruct-GPTQ"
        )

        loras_dir = os.path.join(self.base_dir, "storage", "loras")
        dataset_dir = os.path.join(self.base_dir, "storage", "datasets", name)

        os.makedirs(loras_dir, exist_ok=True)
        os.makedirs(dataset_dir, exist_ok=True)

        if stage == "pt":
            for memory in memories:
                item = {
                    "text": memory,
                }
                data.append(item)
        elif stage == "sft":
            if memories:
                with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
                    futures = []
                    llm = byzerllm.ByzerLLM().from_default_model(self.model_name)
                    for memory in memories:
                        futures.append(executor.submit(self.to_qa_pairs, memory, llm))

                    for future in concurrent.futures.as_completed(futures):
                        qa_pairs = future.result()
                        for qa_pair in qa_pairs:
                            item = {
                                "instruction": qa_pair.question,
                                "input": "",
                                "output": qa_pair.answer,
                            }
                            data.append(item)
            else:
                v = read_alpaca_zh()
                data.extend(v)

        if len(data) < min_samples:
            data = data * (min_samples // len(data) + 1)
        
        with open(f"{dataset_dir}/data.json", "w", encoding="utf-8") as f:
            json.dump(data, f, indent=2, ensure_ascii=False)
        
        with open(f"{dataset_dir}/dataset_info.json", "w", encoding="utf-8") as f:
            r = {"data": {"file_name": "data.json"}}
            if stage == "pt":
                r["data"]["columns"] = {"prompt": "text"}
            f.write(
                json.dumps(
                    r,
                    indent=2,
                    ensure_ascii=False,
                )
            )

        args = dict(
            stage="pt",
            do_train=True,
            model_name_or_path=llama_model,
            dataset="data",
            dataset_dir=dataset_dir,
            cutoff_len=1024,
            max_samples=1000000,
            overwrite_cache=True,
            preprocessing_num_workers=1,
            template="llama3",
            finetuning_type="lora",
            lora_target="all",
            output_dir=f"{loras_dir}/{name}",
            overwrite_output_dir=True,
            per_device_train_batch_size=2,
            gradient_accumulation_steps=4,
            lr_scheduler_type="cosine",
            logging_steps=10,
            warmup_ratio=0.1,
            save_steps=1000,
            plot_loss=True,
            learning_rate=5e-5,
            num_train_epochs=1000.0,
            max_grad_norm=1.0,
            quantization_bit=4,
            loraplus_lr_ratio=16.0,
            fp16=True,
            ddp_timeout=180000000,
        )
        os.environ["WANDB_DISABLED"] = "true"
        from llamafactory.train import tuner

        tuner.run_exp({**args, **options})



from transformers import AutoTokenizer, AutoModelForCausalLM
import transformers
import torch
from typing import Dict,List,Tuple


def get_meta(self): 
    config = self.config   
    return [{
        "model_deploy_type": "proprietary",
        "backend":"transformers",
        "max_model_len":getattr(config, "model_max_length", -1),
        "architectures":getattr(config, "architectures", [])
    }]

def stream_chat(self,tokenizer,ins:str, his:List[Tuple[str,str]]=[],  
        max_length:int=4096, 
        top_p:float=0.95,
        temperature:float=0.1,**kwargs):
        
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    tokens = tokenizer(ins, return_token_type_ids=False,return_tensors="pt").to(device)
    response = self.generate(
        input_ids=tokens["input_ids"],
        max_new_tokens=max_length,
        repetition_penalty=1.05,
        temperature=temperature,
        eos_token_id=tokenizer.eos_token_id
    )
    answer = tokenizer.decode(response[0][tokens["input_ids"].shape[1]:], clean_up_tokenization_spaces=False)
    return [(answer,"")]


def init_model(model_dir,infer_params:Dict[str,str]={},sys_conf:Dict[str,str]={}):        
    tokenizer = AutoTokenizer.from_pretrained(model_dir)   
    tokenizer.padding_side="right"
    tokenizer.pad_token_id=0 
    model = AutoModelForCausalLM.from_pretrained(model_dir,trust_remote_code=True,
                                                device_map='auto',                                                
                                                torch_dtype=torch.bfloat16                                                
                                                )
    model.eval()       
    import types
    model.stream_chat = types.MethodType(stream_chat, model)  
    model.get_meta = types.MethodType(get_meta, model)   
    return (model,tokenizer)





import argparse
import shlex
from byzerllm.lang import lang, locales
from typing import List, Optional


class StoreNestedDict(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        d = {}
        for kv in values:
            try:
                for clean_kv in shlex.split(kv):
                    key, value = clean_kv.split("=", 1)
                    d[key] = value
            except ValueError:
                raise argparse.ArgumentError(self, f"Invalid argument format: {kv}")
        setattr(namespace, self.dest, d)


def get_command_args(input_args: Optional[List[str]] = None):
    parser = argparse.ArgumentParser(description=locales["desc"][lang])
    subparsers = parser.add_subparsers(dest="command")

    # Deploy 子命令
    deploy_cmd = subparsers.add_parser("deploy", help=locales["help_deploy"][lang])
    deploy_cmd.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    deploy_cmd.add_argument(
        "--num_workers", type=int, default=1, help=locales["help_num_workers"][lang]
    )
    deploy_cmd.add_argument(
        "--worker_concurrency",
        type=int,
        default=1,
        help=locales["help_worker_concurrency"][lang],
    )
    deploy_cmd.add_argument(
        "--gpus_per_worker",
        type=float,
        default=0,
        help=locales["help_gpus_per_worker"][lang],
    )
    deploy_cmd.add_argument(
        "--cpus_per_worker",
        type=float,
        default=0.01,
        help=locales["help_cpus_per_worker"][lang],
    )
    deploy_cmd.add_argument(
        "--model_path",
        default="",
        required=False,
        help=locales["help_model_path"][lang],
    )
    deploy_cmd.add_argument(
        "--pretrained_model_type",
        default="custom/llama2",
        help=locales["help_pretrained_model_type"][lang],
    )
    deploy_cmd.add_argument(
        "--model", required=True, help=locales["help_udf_name"][lang]
    )
    deploy_cmd.add_argument(
        "--infer_backend", default="", help=locales["help_infer_backend"][lang]
    )
    deploy_cmd.add_argument(
        "--infer_params",
        nargs="+",
        action=StoreNestedDict,
        help=locales["help_infer_params"][lang],
    )
    deploy_cmd.add_argument("--file", default=None, required=False, help="")

    # Undeploy 子命令
    deploy_cmd = subparsers.add_parser("undeploy", help=locales["help_deploy"][lang])
    deploy_cmd.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    deploy_cmd.add_argument(
        "--model", required=True, help=locales["help_udf_name"][lang]
    )
    deploy_cmd.add_argument("--file", default=None, required=False, help="")
    deploy_cmd.add_argument("--force", action="store_true", help="")

    # Query 子命令
    query_cmd = subparsers.add_parser("query", help=locales["help_query"][lang])
    query_cmd.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    query_cmd.add_argument(
        "--model", required=True, help=locales["help_query_model"][lang]
    )
    query_cmd.add_argument(
        "--query", required=True, help=locales["help_query_text"][lang]
    )
    query_cmd.add_argument(
        "--template", default="auto", help=locales["help_template"][lang]
    )
    query_cmd.add_argument("--file", default=None, required=False, help="")
    query_cmd.add_argument("--output_file", default="", help="")

    # Stat Command
    status_cmd = subparsers.add_parser("stat", help=locales["help_query"][lang])
    status_cmd.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    status_cmd.add_argument(
        "--model", required=True, help=locales["help_query_model"][lang]
    )
    status_cmd.add_argument("--file", default=None, required=False, help="")

    # Storage 子命令
    storage_cmd = subparsers.add_parser("storage", help="Manage Byzer Storage")
    storage_cmd_subparsers = storage_cmd.add_subparsers(dest="storage_command")

    emb_parser = storage_cmd_subparsers.add_parser("emb", help="Manage embedding model")
    emb_subparsers = emb_parser.add_subparsers(dest="emb_command", required=True)
    emb_start_parser = emb_subparsers.add_parser("start", help="Start embedding model")
    emb_stop_parser = emb_subparsers.add_parser("stop", help="Stop embedding model")

    model_memory_parser = storage_cmd_subparsers.add_parser(
        "model_memory", help="Manage long-term memory model"
    )
    model_memory_subparsers = model_memory_parser.add_subparsers(
        dest="model_memory_command", required=True
    )
    model_memory_start_parser = model_memory_subparsers.add_parser(
        "start", help="Start long-term memory model"
    )
    model_memory_stop_parser = model_memory_subparsers.add_parser(
        "stop", help="Stop long-term memory model"
    )

    # Add common arguments for all new subcommands
    for _subparser in [
        emb_start_parser,
        emb_stop_parser,
        model_memory_start_parser,
        model_memory_stop_parser,
    ]:
        _subparser.add_argument("--file", default=None, required=False, help="")
        _subparser.add_argument(
            "--ray_address", default="auto", help=locales["help_ray_address"][lang]
        )
        _subparser.add_argument("--version", default="0.1.11", required=False, help="")
        _subparser.add_argument("--cluster", default="byzerai_store", help="")
        _subparser.add_argument("--base_dir", default="", help="")

    # install 子命令
    storage_install_cmd = storage_cmd_subparsers.add_parser(
        "install", help="Install Byzer Storage"
    )
    storage_install_cmd.add_argument("--file", default=None, required=False, help="")
    storage_install_cmd.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    storage_install_cmd.add_argument(
        "--version", default="0.1.11", required=False, help=""
    )
    storage_install_cmd.add_argument("--cluster", default="byzerai_store", help="")
    storage_install_cmd.add_argument("--base_dir", default="", help="")

    storage_collection_cmd = storage_cmd_subparsers.add_parser("collection", help="")
    storage_collection_cmd.add_argument("--file", default=None, required=False, help="")
    storage_collection_cmd.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    storage_collection_cmd.add_argument(
        "--version", default="0.1.11", required=False, help=""
    )
    storage_collection_cmd.add_argument("--cluster", default="byzerai_store", help="")
    storage_collection_cmd.add_argument("--base_dir", default="", help="")
    storage_collection_cmd.add_argument("--name", default="", help="")
    storage_collection_cmd.add_argument("--description", default="", help="")

    # start 子命令
    storage_start_command = storage_cmd_subparsers.add_parser(
        "start", help="Start Byzer Storage"
    )
    storage_start_command.add_argument("--file", default=None, required=False, help="")
    storage_start_command.add_argument(
        "--version", default="0.1.11", required=False, help=""
    )
    storage_start_command.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    storage_start_command.add_argument("--cluster", default="byzerai_store", help="")
    storage_start_command.add_argument("--base_dir", default="", help="")
    storage_start_command.add_argument(
        "--enable_emb", action="store_true", help="Enable embedding model"
    )    
    storage_start_command.add_argument(
        "--enable_model_memory",
        action="store_true",
        help="Enable model memory",
    )
    storage_start_command.add_argument(
        "--num_nodes",
        type=int,
        default=1,
        help="Number of nodes in the cluster",
    )
    storage_start_command.add_argument(
        "--node_cpus",
        type=int,
        default=1,
        help="Number of CPUs per node",
    )
    storage_start_command.add_argument(
        "--node_memory",
        type=int,
        default=2,
        help="Memory per node in GB",
    )

    # stop 子命令
    storage_stop_command = storage_cmd_subparsers.add_parser(
        "stop", help="Stop Byzer Storage"
    )
    storage_stop_command.add_argument("--file", default=None, required=False, help="")
    storage_stop_command.add_argument(
        "--version", default="0.1.11", required=False, help=""
    )
    storage_stop_command.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    storage_stop_command.add_argument("--cluster", default="byzerai_store", help="")
    storage_stop_command.add_argument("--base_dir", default="", help="")

    # export 子命令
    storage_export_command = storage_cmd_subparsers.add_parser(
        "export", help="Export Byzer Storage Information"
    )
    storage_export_command.add_argument("--file", default=None, required=False, help="")
    storage_export_command.add_argument(
        "--version", default="0.1.11", required=False, help=""
    )
    storage_export_command.add_argument(
        "--ray_address", default="auto", help=locales["help_ray_address"][lang]
    )
    storage_export_command.add_argument("--cluster", default="byzerai_store", help="")
    storage_export_command.add_argument("--base_dir", default="", help="")

    # Serve 子命令
    serve_cmd = subparsers.add_parser(
        "serve", help="Serve deployed models with OpenAI compatible APIs"
    )
    serve_cmd.add_argument("--file", default="", help="")
    serve_cmd.add_argument("--ray_address", default="auto", help="")
    serve_cmd.add_argument("--host", default="", help="")
    serve_cmd.add_argument("--port", type=int, default=8000, help="")
    serve_cmd.add_argument("--uvicorn_log_level", default="info", help="")
    serve_cmd.add_argument("--allow_credentials", action="store_true", help="")
    serve_cmd.add_argument("--allowed_origins", default=["*"], help="")
    serve_cmd.add_argument("--allowed_methods", default=["*"], help="")
    serve_cmd.add_argument("--allowed_headers", default=["*"], help="")
    serve_cmd.add_argument("--api_key", default="", help="")
    serve_cmd.add_argument("--served_model_name", default="", help="")
    serve_cmd.add_argument("--prompt_template", default="", help="")
    serve_cmd.add_argument("--ssl_keyfile", default="", help="")
    serve_cmd.add_argument("--ssl_certfile", default="", help="")
    serve_cmd.add_argument("--response_role", default="assistant", help="")
    serve_cmd.add_argument(
        "--template", default="auto", help=locales["help_template"][lang]
    )

    if input_args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(input_args)
    return args



from langchain.vectorstores import FAISS
from langchain.docstore.document import Document
from pathlib import Path
from typing import List



from . import BuilderParams
from .builder import OnceWay,MergeWay
from .client import ByzerLLMClient,LocalEmbeddings

class VectorDB:
    def __init__(self,db_dir:str,client:ByzerLLMClient,extra_params={}) -> None:
        self.db_dir = db_dir 
        self.db = None  
        self.client = client 
        self.embeddings = LocalEmbeddings(self.client,extra_params.get("promptPrefix",None))     
    
    def _is_visible(self,p: Path) -> bool:
        parts = p.parts
        for _p in parts:
            if _p.startswith("."):
                return False
        return True

    
    def save(self,path,params:BuilderParams,extra_params={}):                        
        if params.batch_size == 0:
            b = OnceWay(self.db_dir,self.embeddings)
            b.build(path,params,extra_params)
        else:
            b = MergeWay(self.db_dir,self.embeddings)
            b.build(path,params,extra_params)    
            

    def build_from(self,path,params:BuilderParams,extra_params={}):
        return self.save(path,params,extra_params) 


    def merge_from(self,target_path:str):
        if self.db is None:
            self.db = FAISS.from_documents([], self.embeddings)            

        self.db.merge_from(FAISS.load_local(target_path,self.embeddings))                

    def query(self,prompt:str,s:str,k=4)->List[Document]:
        if not self.db:
           self.db = FAISS.load_local(self.db_dir,self.embeddings)        
        result = self.db.similarity_search_with_score(prompt + s,k=k)
        return result


from transformers import AutoModelForCausalLM, AutoTokenizer,GenerationConfig
import json
import os
import io
from typing import Any,Any,Dict, List,Tuple,Generator
import base64
import uuid
import tempfile

def get_meta(self): 
    config = self.config   
    return [{
        "model_deploy_type": "proprietary",
        "backend":"transformers",
        "max_model_len":getattr(config, "model_max_length", -1),
        "architectures":getattr(config, "architectures", []),
        "message_format":True,
    }]

def stream_chat(self,tokenizer,ins:str, his:List[Dict[str,str]]=[],  
        max_length:int=4096, 
        top_p:float=0.95,
        temperature:float=0.1,**kwargs):
    image = kwargs["image"]
    image_b = base64.b64decode(image)

    temp_image_dir = os.path.join(tempfile.gettempdir(),"byzerllm","visualglm","images")
    if "temp_image_dir" in kwargs:
        temp_image_dir = kwargs["temp_image_dir"]

    if not os.path.exists(temp_image_dir):
        os.makedirs(temp_image_dir)

    image_file = os.path.join(temp_image_dir,f"{str(uuid.uuid4())}.jpg")
    
    if "input_image_path" in kwargs:
        image_file = kwargs["input_image_path"]

    with open(image_file,"wb") as f:
        f.write(image_b)

    # history format [('Picture 1:<img>https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen-VL/assets/demo.jpeg</img>\n这是什么?',
    # '图中是一名女子和她的狗在沙滩上玩耍，狗的品种应该是一只拉布拉多犬。')]     
    input_history = None
    if "history" in kwargs:
        input_history = []
        temp_input_history = json.loads(kwargs["history"]) 
        temp_input_history_length = len(temp_input_history)
        for i in range(0,temp_input_history_length,2):
            input_history.append((temp_input_history[i]["content"],temp_input_history[i+1]["content"]))
            
    if not input_history:           
        query = tokenizer.from_list_format([
        {'image': image_file}, 
        {'text': ins},])
        response, history = self.chat(tokenizer, query=query, history=None)                            
    else:        
        response, history = self.chat(tokenizer, ins, history=input_history)                            
    
    temp_history = []
    for item in history:
        temp_history.append(
            {"role":"user","content":item[0]},
        )
        temp_history.append(
            {"role":"assistant","content":item[1]},
        )
        
    output = json.dumps({"response":response,"history":temp_history},ensure_ascii=False)
    return [(output,{"metadata":{}})] 


def init_model(model_dir,infer_params:Dict[str,str]={},sys_conf:Dict[str,str]={}): 
    pretrained_model_dir = os.path.join(model_dir,"pretrained_model")
    adaptor_model_dir = model_dir
    is_adaptor_model = os.path.exists(pretrained_model_dir)
    
    if not is_adaptor_model:        
        pretrained_model_dir = model_dir
    

    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_dir,trust_remote_code=True)    
    model = AutoModelForCausalLM.from_pretrained(pretrained_model_dir,trust_remote_code=True,                                                
                                                bf16=True,
                                                device_map='auto'
                                                ).half().cuda()
    if is_adaptor_model:
        from peft import PeftModel
        model = PeftModel.from_pretrained(model, adaptor_model_dir)
        
    model.eval()
    model.generation_config = GenerationConfig.from_pretrained(pretrained_model_dir, trust_remote_code=True)       
    import types
    model.stream_chat = types.MethodType(stream_chat, model)     
    model.get_meta = types.MethodType(get_meta, model)
    return (model,tokenizer)









from byzerllm.apps.agent import Agents
from byzerllm.apps.agent.user_proxy_agent import UserProxyAgent
from byzerllm.apps.agent.extensions.data_analysis import DataAnalysis
from byzerllm.utils.client import ByzerLLM,Templates
import os

chat_model_name = "qianwen_chat"
llm = ByzerLLM()
llm.setup_default_model_name(chat_model_name) 
llm.setup_template(chat_model_name,template=Templates.qwen())

current_file_directory = os.path.dirname(os.path.abspath(__file__))


def test_get_agent_names():
    data_analysis = DataAnalysis("chat4","william",os.path.join(current_file_directory,"test.csv"),
                             llm,None,skip_preview_file=True)
    
    v = data_analysis.get_agent_names()
    assert set(v) == set(['assistant_agent', 'visualization_agent', 'common_agent', 'privew_file_agent', 'python_interpreter'])
    data_analysis.close()

def test_update_pipeline_system_message():    
    data_analysis = DataAnalysis("chat4","william",os.path.join(current_file_directory,"test.csv"),
                             llm,None,skip_preview_file=True)
    
    data_analysis.update_pipeline_system_message("hello")
    assert data_analysis.get_pipeline_system_message() == "hello"
    data_analysis.close()

def test_system_message_by_agent():
    data_analysis = DataAnalysis("chat4","william",os.path.join(current_file_directory,"test.csv"),
                             llm,None,skip_preview_file=True)
    
    data_analysis.update_agent_system_message("assistant_agent","hello")
    assert data_analysis.get_agent_system_message("assistant_agent") == "hello"
    data_analysis.close()


import json
from torch.utils.data import Dataset


class SFTDataset(Dataset):
    def __init__(self, file, tokenizer, max_seq_length,**kwargs):
        self.tokenizer = tokenizer
        self.bos_token_id = tokenizer.bos_token_id
        self.eos_token_id = tokenizer.eos_token_id

        if self.bos_token_id is None and self.eos_token_id is None:
            print("bos_token_id or eos_token_id are both unset",flush=True)

        self.eos_token = tokenizer.eos_token
        self.bos_token = tokenizer.bos_token
        self.max_seq_length = max_seq_length
        print('Loading data: {}'.format(file))
        with open(file, 'r', encoding='utf8') as f:
            data_list = f.readlines()
        print("there are {} data in dataset".format(len(data_list)))
        self.data_list = data_list
        self.system_msg = kwargs.get('system_msg', 'You are a helpful assistant. Think it over and answer the user question correctly.\n')
        self.user_role = kwargs.get('user_role', 'User')
        self.assistant_role = kwargs.get('assistant_role', 'Assistant')
        self.user_role_prefix = f"{self.user_role}:"
        self.assistant_role_prefix = f"{self.assistant_role}:"

        self.dataset_tokens_count = 0

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, index):
                
        data = self.data_list[index]
        data = json.loads(data)
        conversation = data['conversation']

        # 收集多轮对话
        utterances = [self.system_msg]
        for x in conversation:
            utterances.append(f"{self.user_role_prefix}{x['human']}\n")
            utterances.append(f"{self.assistant_role_prefix}:{x['assistant']}\n")
        utterances_ids = self.tokenizer(utterances).input_ids

        # 模型的输入格式为：<s>User:input1</s>target1</s>input2</s>target2</s>...
        if self.bos_token_id is None:
            input_ids = []
            target_mask = []  # 用于对input进行mask，只计算target部分的loss           
        else:
            input_ids = [self.bos_token_id]
            target_mask = [0]  # 用于对input进行mask，只计算target部分的loss

        if self.eos_token_id is None:
            input_ids_suffix   = []
            target_mask_suffix = 0
        else:
            input_ids_suffix   = [self.eos_token_id]
            target_mask_suffix = 1
        
        for i, utterances_id in enumerate(utterances_ids):
            input_ids += (utterances_id + input_ids_suffix)
            if i % 2 == 0:
                target_mask += [0] * (len(utterances_id) + target_mask_suffix)
            else:
                target_mask += [1] * (len(utterances_id) + target_mask_suffix)

        assert len(input_ids) == len(target_mask)
        # 对长度进行截断
        input_ids = input_ids[:self.max_seq_length]
        target_mask = target_mask[:self.max_seq_length]
        attention_mask = [1] * len(input_ids)
        assert len(input_ids) == len(target_mask) == len(attention_mask)

        self.dataset_tokens_count += len(input_ids)

        inputs = {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'target_mask': target_mask
        }
        return inputs



import pydantic
from typing import Optional, List, Dict, Any
from enum import Enum

class Action(Enum):
    CONTINUE = "continue"
    STOP = "stop"    

class QueryRewriteResult(pydantic.BaseModel):
    message: Dict[str, Any]    
    action: Action
    extra_info: Dict[str, Any]    


import os
import json
from typing import Optional
from dataclasses import dataclass, field


CHATGLM_REPO_NAME = "THUDM/chatglm-6b"
CHATGLM_LASTEST_HASH = "35ca52301fbedee885b0838da5d15b7b47faa37c"


@dataclass
class DatasetAttr:

    load_from: str
    dataset_name: Optional[str] = None
    file_name: Optional[str] = None
    file_sha1: Optional[str] = None

    def __post_init__(self):
        self.prompt_column = "instruction"
        self.query_column = "input"
        self.response_column = "output"
        self.history_column = None


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune.
    """
    model_name_or_path: Optional[str] = field(
        default=CHATGLM_REPO_NAME,
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models."}
    )
    config_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained config name or path if not the same as model_name."}
    )
    tokenizer_name: Optional[str] = field(
        default=None,
        metadata={"help": "Pretrained tokenizer name or path if not the same as model_name."}
    )
    cache_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Where to store the pretrained models downloaded from huggingface.co."}
    )
    use_fast_tokenizer: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."}
    )
    model_revision: Optional[str] = field(
        default=CHATGLM_LASTEST_HASH,
        metadata={"help": "The specific model version to use (can be a branch name, tag name or commit id)."}
    )
    use_auth_token: Optional[bool] = field(
        default=False,
        metadata={"help": "Will use the token generated when running `huggingface-cli login`."}
    )
    quantization_bit: Optional[int] = field(
        default=None,
        metadata={"help": "The number of bits to quantize the model."}
    )
    checkpoint_dir: Optional[str] = field(
        default=None,
        metadata={"help": "Path to the directory containing the model checkpoints as well as the configurations."}
    )

    def __post_init__(self):
        if self.checkpoint_dir is not None: # support merging lora weights
            self.checkpoint_dir = [cd.strip() for cd in self.checkpoint_dir.split(",")]


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and evaluation.
    """
    dataset: Optional[str] = field(
        default="alpaca_zh",
        metadata={"help": "The name of provided dataset(s) to use. Use comma to separate multiple datasets."}
    )
    dataset_dir: Optional[str] = field(
        default="data",
        metadata={"help": "The name of the folder containing datasets."}
    )
    split: Optional[str] = field(
        default="train",
        metadata={"help": "Which dataset split to use for training and evaluation."}
    )
    overwrite_cache: Optional[bool] = field(
        default=False,
        metadata={"help": "Overwrite the cached training and evaluation sets."}
    )
    preprocessing_num_workers: Optional[int] = field(
        default=None,
        metadata={"help": "The number of processes to use for the preprocessing."}
    )
    max_source_length: Optional[int] = field(
        default=512,
        metadata={"help": "The maximum total input sequence length after tokenization."}
    )
    max_target_length: Optional[int] = field(
        default=512,
        metadata={"help": "The maximum total output sequence length after tokenization."}
    )
    max_samples: Optional[int] = field(
        default=None,
        metadata={"help": "For debugging purposes, truncate the number of examples for each dataset."}
    )
    num_beams: Optional[int] = field(
        default=None,
        metadata={"help": "Number of beams to use for evaluation. This argument will be passed to `model.generate`"}
    )
    ignore_pad_token_for_loss: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to ignore the tokens corresponding to padded labels in the loss computation or not."}
    )
    source_prefix: Optional[str] = field(
        default=None,
        metadata={"help": "A prefix to add before every source text (useful for T5 models)."}
    )

    def __post_init__(self): # support mixing multiple datasets        
        self.dataset_list = []

        dataset_attr = DatasetAttr(
                    "file",
                    file_name="data.jsonl",
                    file_sha1= None
                )
        self.dataset_list.append(dataset_attr)        
        


@dataclass
class FinetuningArguments:
    """
    Arguments pertaining to which techniques we are going to fine-tuning with.
    """
    finetuning_type: Optional[str] = field(
        default="lora",
        metadata={"help": "Which fine-tuning method to use."}
    )
    num_layer_trainable: Optional[int] = field(
        default=3,
        metadata={"help": "Number of trainable layers for Freeze fine-tuning."}
    )
    name_module_trainable: Optional[str] = field(
        default="mlp",
        metadata={"help": "Name of trainable modules for Freeze fine-tuning."}
    )
    pre_seq_len: Optional[int] = field(
        default=16,
        metadata={"help": "Number of prefix tokens to use for P-tuning V2."}
    )
    prefix_projection: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to add a project layer for the prefix in P-tuning V2 or not."}
    )
    lora_rank: Optional[int] = field(
        default=8,
        metadata={"help": "The intrinsic dimension for LoRA fine-tuning."}
    )
    lora_alpha: Optional[float] = field(
        default=32.0,
        metadata={"help": "The scale factor for LoRA fine-tuning. (similar with the learning rate)"}
    )
    lora_dropout: Optional[float] = field(
        default=0.1,
        metadata={"help": "Dropout rate for the LoRA fine-tuning."}
    )
    lora_target: Optional[str] = field(
        default="query_key_value",
        metadata={"help": "Name(s) of target modules to apply LoRA. Use comma to separate multiple modules."}
    )
    resume_lora_training: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to resume training from the last LoRA weights or create new weights after merging them."}
    )
    plot_loss: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to plot the training loss after fine-tuning or not."}
    )

    def __post_init__(self):
        self.lora_target = [target.strip() for target in self.lora_target.split(",")] # support custom target modules of LoRA

        if self.num_layer_trainable > 0: # fine-tuning the last n layers if num_layer_trainable > 0
            trainable_layer_ids = [27-k for k in range(self.num_layer_trainable)]
        else: # fine-tuning the first n layers if num_layer_trainable < 0
            trainable_layer_ids = [k for k in range(-self.num_layer_trainable)]
        if self.name_module_trainable == "mlp":
            self.trainable_layers = ["layers.{:d}.mlp".format(idx) for idx in trainable_layer_ids]
        elif self.name_module_trainable == "qkv":
            self.trainable_layers = ["layers.{:d}.attention.query_key_value".format(idx) for idx in trainable_layer_ids]

        if self.finetuning_type not in ["none", "freeze", "p_tuning", "lora"]:
            raise NotImplementedError("Invalid fine-tuning method.")



from byzerllm.utils.client import ByzerLLM,InferBackend,Templates,LLMRequest
import ray
import pytest

model_path = "/home/winubuntu/models/Qwen-1_8B-Chat"

class TestByzerLLMVLLMDeploy(object):
    llm = None

    def setup_class(self):
        if ray.is_initialized():
            ray.shutdown()
        
        ray.init(address="auto",namespace="default")
        
        self.llm = ByzerLLM(verbose=True)

        if self.llm.is_model_exist("chat"):
            self.llm.undeploy("chat")

        self.llm.setup_gpus_per_worker(1).setup_num_workers(1).setup_infer_backend(InferBackend.VLLM)
        self.llm.deploy(
            model_path=model_path,
            pretrained_model_type="custom/auto",
            udf_name="chat",
            infer_params={}
        )
        self.llm.setup_default_model_name("chat")
        self.llm.setup_template("chat",Templates.qwen())
        
        
    def teardown_class(self):
        if self.llm.is_model_exist("chat"):
            self.llm.undeploy("chat")
    
    def test_get_meta(self):
        t = self.llm.get_meta("chat")
        assert t["model_deploy_type"] == "proprietary"
        assert t["backend"] == "ray/vllm"



    
